import strawberry
import structlog
from redis.asyncio import Redis as AIORedis

from oauth2_lib.strawberry import authenticated_mutation_field
from orchestrator.api.api_v1.endpoints.settings import generate_engine_status_response
from orchestrator.graphql.resolvers.helpers import make_async
from orchestrator.graphql.schemas.errors import Error
from orchestrator.graphql.schemas.settings import (
    CACHE_FLUSH_OPTIONS,
    CacheClearResponse,
    CacheClearSuccess,
    EngineSettingsType,
    StatusType,
    StatusUpdateResponse,
    WorkerStatusType,
)
from orchestrator.graphql.types import OrchestratorInfo
from orchestrator.graphql.utils import get_selected_fields
from orchestrator.schemas.engine_settings import EngineSettingsSchema, WorkerStatus
from orchestrator.services.processes import SYSTEM_USER, ThreadPoolWorkerStatus, marshall_processes
from orchestrator.services.settings import get_engine_settings, get_engine_settings_for_update, post_update_to_slack
from orchestrator.settings import ExecutorType, app_settings
from orchestrator.utils.redis import delete_keys_matching_pattern
from orchestrator.utils.redis_client import create_redis_asyncio_client

logger = structlog.get_logger(__name__)


# Queries
@make_async
def resolve_settings(info: OrchestratorInfo) -> StatusType:
    selected_fields = get_selected_fields(info)

    db_engine_settings = get_engine_settings()
    pydantic_orm_resp = generate_engine_status_response(db_engine_settings)
    engine_settings = EngineSettingsType.from_pydantic(pydantic_orm_resp)

    settings_resp_obj = StatusType(
        engine_settings=engine_settings,
        cache_names=None,
        worker_status=None,
    )

    if "workerStatus" in selected_fields:
        worker_status: WorkerStatus
        if app_settings.EXECUTOR == ExecutorType.WORKER:
            from orchestrator.services.tasks import CeleryJobWorkerStatus

            worker_status = CeleryJobWorkerStatus()
        else:
            worker_status = ThreadPoolWorkerStatus()
        settings_resp_obj.worker_status = WorkerStatusType.from_pydantic(worker_status)

    if "cacheNames" in selected_fields:
        settings_resp_obj.cache_names = CACHE_FLUSH_OPTIONS

    return settings_resp_obj


# Mutations
async def clear_cache(info: OrchestratorInfo, name: str) -> CacheClearSuccess | Error:
    cache: AIORedis = create_redis_asyncio_client(app_settings.CACHE_URI.get_secret_value())
    if name not in CACHE_FLUSH_OPTIONS:
        return Error(message="Invalid cache name")

    key_name = "orchestrator:*" if name == "all" else f"orchestrator:{name}:*"
    deleted = await delete_keys_matching_pattern(cache, key_name)
    return CacheClearSuccess(deleted=deleted)


async def set_status(info: OrchestratorInfo, global_lock: bool) -> Error | EngineSettingsType:
    engine_settings = get_engine_settings_for_update()

    result = marshall_processes(engine_settings, global_lock)
    if not result:
        return Error(
            message="Something went wrong while updating the database aborting, possible manual intervention required",
        )
    if app_settings.SLACK_ENGINE_SETTINGS_HOOK_ENABLED:
        oidc_user = await info.context.get_current_user
        user_name = oidc_user.name if oidc_user else SYSTEM_USER
        post_update_to_slack(EngineSettingsSchema.model_validate(result), user_name)

    status_response = generate_engine_status_response(result)
    return EngineSettingsType.from_pydantic(status_response)


@strawberry.type(description="Settings endpoint mutations")
class SettingsMutation:
    clear_cache: CacheClearResponse = authenticated_mutation_field(
        resolver=clear_cache, description="Clear a redis cache by name"
    )
    update_status: StatusUpdateResponse = authenticated_mutation_field(
        resolver=set_status, description="Update global status of the engine"
    )



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

warnings.warn(
    "ALL_SCHEDULERS is deprecated and will be removed in 5.0.0. "
    "Scheduling tasks can now be handled entirely through the API. "
    "For more details, please consult https://workfloworchestrator.org/orchestrator-core/guides/upgrading/4.7/",
    DeprecationWarning,
    stacklevel=2,
)
ALL_SCHEDULERS: list = []



# Copyright 2019-2025 SURF, GÉANT, ESnet.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import contextvars
import functools
import inspect
import secrets
from collections.abc import Callable
from dataclasses import asdict, dataclass
from itertools import dropwhile
from typing import (
    Any,
    Generic,
    NoReturn,
    Protocol,
    TypeVar,
    cast,
    overload,
    runtime_checkable,
)
from uuid import UUID

import strawberry
import structlog
from structlog.contextvars import bound_contextvars
from structlog.stdlib import BoundLogger

from nwastdlib import const, identity
from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.config.assignee import Assignee
from orchestrator.db import db, transactional
from orchestrator.services.settings import get_engine_settings
from orchestrator.targets import Target
from orchestrator.types import ErrorDict, StepFunc
from orchestrator.utils.auth import Authorizer
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.docs import make_workflow_doc
from orchestrator.utils.errors import error_state_to_dict
from orchestrator.utils.state import form_inject_args, inject_args
from pydantic_forms.core import FormPage
from pydantic_forms.types import (
    FormGenerator,
    InputFormGenerator,
    InputStepFunc,
    State,
    StateInputFormGenerator,
    StateInputStepFunc,
    StateSimpleInputFormGenerator,
    strEnum,
)

logger = structlog.get_logger(__name__)

StepLogFunc = Callable[["ProcessStat", "Step", "Process"], "Process"]
StepLogFuncInternal = Callable[["Step", "Process"], "Process"]
StepToProcessFunc = Callable[[State], "Process"]

step_log_fn_var: contextvars.ContextVar[StepLogFuncInternal] = contextvars.ContextVar("log_step_fn")

DEFAULT_CALLBACK_ROUTE_KEY = "callback_route"
CALLBACK_TOKEN_KEY = "__callback_token"  # noqa: S105
DEFAULT_CALLBACK_PROGRESS_KEY = "callback_progress"  # noqa: S105


@runtime_checkable
class Step(Protocol):
    __name__: str
    __qualname__: str
    name: str
    form: InputFormGenerator | None
    assignee: Assignee | None
    resume_auth_callback: Authorizer | None = None
    retry_auth_callback: Authorizer | None = None

    def __call__(self, state: State) -> Process: ...


@dataclass(frozen=True)
class RunPredicatePass:
    """Return this from a run predicate to allow the workflow to start."""

    def __bool__(self) -> bool:
        return True


@dataclass(frozen=True)
class RunPredicateFail:
    """Return this from a run predicate to block the workflow from starting."""

    message: str

    def __bool__(self) -> bool:
        return False


RunPredicateResult = RunPredicatePass | RunPredicateFail
RunPredicate = Callable[..., RunPredicateResult]


@runtime_checkable
class Workflow(Protocol):
    __name__: str
    __qualname__: str
    name: str
    description: str
    authorize_callback: Authorizer
    retry_auth_callback: Authorizer
    initial_input_form: InputFormGenerator | None = None
    target: Target
    steps: StepList
    run_predicate: RunPredicate | None = None

    def __call__(self) -> NoReturn: ...


@dataclass(frozen=True)
class PredicateContext:
    """Context passed as the first argument to a workflow's run predicate."""

    workflow: Workflow
    workflow_key: str


def make_step_function(
    f: Callable,
    name: str,
    form: InputFormGenerator | None = None,
    assignee: Assignee | None = Assignee.SYSTEM,
    resume_auth_callback: Authorizer | None = None,
    retry_auth_callback: Authorizer | None = None,
) -> Step:
    step_func = cast(Step, f)

    step_func.name = name
    step_func.form = form
    step_func.assignee = assignee
    step_func.resume_auth_callback = resume_auth_callback
    step_func.retry_auth_callback = retry_auth_callback
    return step_func


class StepList(list[Step]):
    """Wraps around a primitive list of `Step` to provide a "list" with associative `append` (or its alias: `>>`).

    >>> one = step("one")(dict)
    >>> two = step("two")(dict)
    >>> three = step("three")(dict)

    >>> empty = StepList([])
    >>> empty >> empty
    StepList []

    >>> empty == empty >> empty
    True

    >>> str(empty >> one)
    'StepList [one]'

    >>> (begin >> one >> two) >> three == begin >> one >> (begin >> two >> three)
    True
    """

    def map(self, f: Callable) -> StepList:
        return StepList(map(f, self))

    @overload  # type: ignore
    def __getitem__(self, i: int) -> Step: ...

    @overload
    def __getitem__(self, i: slice) -> StepList: ...

    def __getitem__(self, i: int | slice) -> Step | StepList:
        retval: Step | list[Step] = super().__getitem__(i)
        if isinstance(retval, list):
            # ensure we return a StepList and not a regular list.
            retval = type(self)(retval)
        return retval

    def __rshift__(self, other: StepList | Step) -> StepList:
        if isinstance(other, Step):
            return StepList([*self, other])

        if isinstance(other, StepList):
            return StepList([*self, *other])

        if hasattr(other, "__name__"):  # type:ignore
            raise ValueError(
                f"Expected @step decorated function or type Step or StepList, got {type(other)} with name {other.__name__} instead."
            )
        raise ValueError(f"Expected @step decorated function or type Step or StepList, got {type(other)} instead.")

    def __str__(self) -> str:
        return f"StepList [{', '.join(x.name for x in self)}]"

    def __repr__(self) -> str:
        return f"StepList [{', '.join(repr(x) for x in self)}]"


def _handle_simple_input_form_generator(f: StateInputStepFunc) -> StateInputFormGenerator:
    """Processes f into a form generator and injects a pre-hook for user authorization."""
    if inspect.isgeneratorfunction(f):
        return cast(StateInputFormGenerator, f)
    if inspect.isgenerator(f):
        raise ValueError("Got a generator object instead of function, this is not correct")

    # If f is a SimpleInputFormGenerator convert to new style generator function
    def form_generator(state: State) -> FormGenerator:
        user_input: FormPage = yield cast(StateSimpleInputFormGenerator, f)(state)
        return user_input.model_dump()

    return form_generator


async def allow(_: OIDCUserModel | None) -> bool:
    """Default function to return True in absence of user-defined authorize function."""
    return True


def make_workflow(
    f: Callable,
    description: str,
    initial_input_form: InputStepFunc | None,
    target: Target,
    steps: StepList,
    authorize_callback: Authorizer | None = None,
    retry_auth_callback: Authorizer | None = None,
    run_predicate: RunPredicate | None = None,
) -> Workflow:
    @functools.wraps(f)
    def wrapping_function() -> NoReturn:
        raise Exception("This function should not be executed")

    wrapping_function = cast(Workflow, wrapping_function)

    wrapping_function.name = f.__name__  # default, will be changed by LazyWorkflowInstance
    wrapping_function.description = description
    wrapping_function.authorize_callback = allow if authorize_callback is None else authorize_callback
    # If no retry auth policy is given, defer to policy for process creation.
    wrapping_function.retry_auth_callback = (
        wrapping_function.authorize_callback if retry_auth_callback is None else retry_auth_callback
    )

    if initial_input_form is None:
        # We always need a form to prevent starting a workflow when no input is needed.
        # This would happen on first post that is used to retrieve the first form page
        initial_input_form = cast(InputStepFunc, const(FormPage))

    wrapping_function.initial_input_form = _handle_simple_input_form_generator(initial_input_form)
    wrapping_function.target = target
    wrapping_function.steps = steps
    wrapping_function.run_predicate = run_predicate

    wrapping_function.__doc__ = make_workflow_doc(wrapping_function)

    return wrapping_function


def step(
    name: str,
    retry_auth_callback: Authorizer | None = None,
) -> Callable[[StepFunc], Step]:
    """Mark a function as a workflow step."""

    def decorator(func: StepFunc) -> Step:
        @functools.wraps(func)
        def wrapper(state: State) -> Process:
            with bound_contextvars(
                func=func.__qualname__,
                workflow_name=state.get("workflow_name"),
                process_id=state.get("process_id"),
            ):
                step_in_inject_args = inject_args(func)
                try:
                    with transactional(db, logger):
                        result = step_in_inject_args(state)
                        return Success(result)
                except Exception as ex:
                    logger.warning("Step failed", exc_info=ex)
                    return Failed(ex)

        return make_step_function(
            wrapper,
            name,
            retry_auth_callback=retry_auth_callback,
        )

    return decorator


def retrystep(
    name: str,
    retry_auth_callback: Authorizer | None = None,
) -> Callable[[StepFunc], Step]:
    """Mark a function as a retryable workflow step.

    If this step fails it goes to `Waiting` were it will be retried periodically. If it `Success` it acts as a normal
    step.
    """

    def decorator(func: StepFunc) -> Step:
        @functools.wraps(func)
        def wrapper(state: State) -> Process:
            with bound_contextvars(
                func=func.__qualname__,
                workflow_name=state.get("workflow_name"),
                process_id=state.get("process_id"),
            ):
                step_in_inject_args = inject_args(func)
                try:
                    with transactional(db, logger):
                        result = step_in_inject_args(state)
                        return Success(result)
                except Exception as ex:
                    return Waiting(ex)

        return make_step_function(
            wrapper,
            name,
            retry_auth_callback=retry_auth_callback,
        )

    return decorator


def inputstep(
    name: str,
    assignee: Assignee,
    resume_auth_callback: Authorizer | None = None,
    retry_auth_callback: Authorizer | None = None,
) -> Callable[[InputStepFunc], Step]:
    """Add user input step to workflow.

    Any authorization callbacks will be attached to the resulting Step.

    IMPORTANT: In contrast to other workflow steps, the `@inputstep` wrapped function will not run in the
    workflow engine! This means that it must be free of side effects!

    Example::

        @inputstep("User step", assignee=Assignee.NOC)
        def user_step(state: State) -> FormGenerator:
            class Form(FormPage):
                name: str
            user_input = yield Form
            return {**user_input.model_dump(), "some extra key": True}

    """

    def decorator(func: InputStepFunc) -> Step:
        def wrapper(state: State) -> FormGenerator:
            form_generator_in_form_inject_args = form_inject_args(func)

            form_generator = _handle_simple_input_form_generator(form_generator_in_form_inject_args)

            return form_generator(state)

        @functools.wraps(func)
        def suspend(state: State) -> Process:
            return Suspend(state)

        return make_step_function(
            suspend,
            name,
            wrapper,
            assignee,
            resume_auth_callback=resume_auth_callback,
            retry_auth_callback=retry_auth_callback,
        )

    return decorator


def _extend_step_group_steps(name: str, steps: StepList) -> StepList:
    def add_sub_group_info_to_state() -> State:
        return {"__step_name_override": name, "__step_group": name}

    def remove_sub_group_info_from_state() -> State:
        return {"__remove_keys": ["__step_group", "__sub_step"]}

    enter_step = begin >> step(f"{name} - Enter")(add_sub_group_info_to_state)
    exit_step = step(f"{name} - Exit")(remove_sub_group_info_from_state)

    return enter_step >> steps >> exit_step


def step_group(
    name: str, steps: StepList, extract_form: bool = True, retry_auth_callback: Authorizer | None = None
) -> Step:
    """Add a group of steps to the workflow as a single step.

    A step group is a sequence of steps that act as a single step.
    Each step in the step group is a normal step on its own. So they are callables (State) -> Process.
    The state of the group will be the last state of the sub-steps. So if one of the steps goes to FAILED,
    the group goes to failed. If a step goes to SUSPEND, the group is in SUSPEND. If a step goes to SUCCESS however,
    the group is still RUNNING. It will be in SUCCESS only of all the sub-steps are in SUCCESS.

    Args:
        name: The name of the step
        steps: The sub steps in the step group
        extract_form: Whether to attach the first form of the sub steps to the step group
        retry_auth_callback: Callback to determine if user is authorized to retry this group on failure
    """

    steps = _extend_step_group_steps(name, steps)

    def func(initial_state: State) -> Process:
        step_log_fn = step_log_fn_var.get()

        # If sub_step information is present in the state. Resume from the next sub step
        if "__sub_step" in initial_state:
            step_list = StepList(dropwhile(lambda s: s.name != initial_state.get("__sub_step"), steps))[1:]
        else:
            step_list = steps

        def dblogstep(step_: Step, p: Process) -> Process:
            p = p.map(lambda s: s | {"__sub_step": step_.name, "__step_name_override": name})
            # If this is not the first step to be executed, replace previous state
            if step_list[0] != step_ or "__sub_step" in initial_state:
                p = p.map(lambda s: s | {"__replace_last_state": True})
            return step_log_fn(step_, p)

        step_group_start_time = nowtz().timestamp()
        process: Process = Success(initial_state)
        process = _exec_steps(step_list, process, dblogstep)
        # Add instruction to replace state of last sub step before returning process _exec_steps higher in the call tree
        return process.map(
            lambda s: s | {"__replace_last_state": True, "__last_step_started_at": step_group_start_time}
        )

    # Make sure we return a form is a sub step has a form
    form = next((sub_step.form for sub_step in steps if sub_step.form), None) if extract_form else None
    return make_step_function(func, name, form, retry_auth_callback=retry_auth_callback)


def _create_endpoint_step(key: str = DEFAULT_CALLBACK_ROUTE_KEY) -> StepFunc:
    def stepfunc(process_id: UUID) -> State:
        token = secrets.token_urlsafe()
        route = f"/api/processes/{process_id}/callback/{token}"
        # Also add the token under __callback_token for internal use
        return {key: route, CALLBACK_TOKEN_KEY: token}

    return stepfunc


def _awaitstep(name: str, result_key: str | None = None) -> Step:
    def await_(state: State) -> Process:
        if result_key:
            state = {**state, "__callback_result_key": result_key}
        return AwaitingCallback(state)

    return make_step_function(await_, name)


def callback_step(
    name: str,
    action_step: Step,
    validate_step: Step,
    result_key: str | None = None,
    callback_route_key: str = DEFAULT_CALLBACK_ROUTE_KEY,
) -> Step:
    """Creates an asynchronous callback step.

    Internally creates a step group with the following sub steps:

    - Action - This performs the required side effect to an external system. After this, an endpoint is generated and
    stored with the process and the process goes into AWAITING_CALLBACK state.
    - Validate - Uses the provided validate_fn to validate the data coming from the external system.

    The data returned in the callback will be merged in the state. An optional result_key parameter can be supplied
    to specify under which key the data will be merged.
    """
    create_endpoint_step = step(f"{name} - Create endpoint")(_create_endpoint_step(key=callback_route_key))
    await_step = _awaitstep(f"{name} - Await callback", result_key=result_key)
    cleanup_step = step(f"{name} - Cleanup callback step")(lambda: {"__remove_keys": [CALLBACK_TOKEN_KEY]})
    return step_group(
        name=name, steps=begin >> create_endpoint_step >> action_step >> await_step >> validate_step >> cleanup_step
    )


def _purestep(name: str) -> Callable[[StepToProcessFunc], StepList]:
    """Part of workflow "DSL" to map a `state -> Process state` function into a workflow step."""

    def _purestep(f: StepToProcessFunc) -> StepList:
        return StepList([make_step_function(f, name)])

    return _purestep


def conditional(p: Callable[[State], bool]) -> Callable[..., StepList]:
    """Use a predicate to control whether a step is run."""

    def _conditional(steps_or_func: StepList | Step) -> StepList:
        if isinstance(steps_or_func, Step):
            steps = StepList([steps_or_func])
        else:
            steps = steps_or_func

        def wrap(step: Step) -> Step:
            @functools.wraps(step)
            def wrapper(state: State) -> Process:
                return step(state) if p(state) else Skipped(state)

            return make_step_function(wrapper, step.name, step.form, step.assignee)

        return steps.map(wrap)

    return _conditional


def steplens(get: Callable[[State], State], set: Callable[[State], Callable[[State], State]]) -> Callable[[Step], Step]:
    """Update a step list to zoom its input state using get and update its output state using set."""

    def wrap(step: Step) -> Step:
        @functools.wraps(step)
        def wrapper(state: State) -> Process:
            sub_state = get(state)

            result: Process = step(sub_state)

            if result.isfailed() or result.iswaiting():
                return result
            return result.map(set(state))

        return make_step_function(wrapper, step.name, step.form, step.assignee)

    return wrap


def focussteps(key: str) -> Callable[[Step | StepList], StepList]:
    """Return a function that maps `steplens` over `steps`, getting and setting a single key."""

    def zoom(steps_or_func: Step | StepList) -> StepList:
        if isinstance(steps_or_func, Step):
            steps = StepList([steps_or_func])
        else:
            steps = steps_or_func

        def get(state: State) -> State:
            return state.get(key, {})

        def set(state: State) -> Callable[[State], State]:
            return lambda substate: {**state, key: substate}

        return steps.map(steplens(get, set))

    return zoom


def workflow(
    description: str,
    initial_input_form: InputStepFunc | None = None,
    target: Target = Target.SYSTEM,
    authorize_callback: Authorizer | None = None,
    retry_auth_callback: Authorizer | None = None,
    run_predicate: RunPredicate | None = None,
) -> Callable[[Callable[[], StepList]], Workflow]:
    """Transform an initial_input_form and a step list into a workflow.

    Use this for other workflows. For create workflows use :func:`create_workflow`

    Example::

        @workflow("create service port")
        def create_service_port():
            init
            << do_something
            << done
    """
    if initial_input_form is None:
        initial_input_form_in_form_inject_args = None
    else:
        initial_input_form_in_form_inject_args = form_inject_args(initial_input_form)

    def _workflow(f: Callable[[], StepList]) -> Workflow:
        return make_workflow(
            f,
            description,
            initial_input_form_in_form_inject_args,
            target,
            f(),
            authorize_callback=authorize_callback,
            retry_auth_callback=retry_auth_callback,
            run_predicate=run_predicate,
        )

    return _workflow


@dataclass
class ProcessStat:
    process_id: UUID
    workflow: Workflow
    state: Process
    log: StepList  # Remaining steps to execute
    current_user: str
    user_model: OIDCUserModel | None = None

    def update(self, **vs: Any) -> ProcessStat:
        """Update ProcessStat.

        >>> pstat = ProcessStat('', None, {}, [], "")
        >>> pstat.update(state={"a": "b"})
        ProcessStat(process_id='', workflow=None, state={'a': 'b'}, log=[], current_user='', user_model=None)
        """
        return ProcessStat(**{**asdict(self), **vs})


S = TypeVar("S")
F = TypeVar("F")


@strawberry.enum
class ProcessStatus(strEnum):
    CREATED = "created"
    RUNNING = "running"
    SUSPENDED = "suspended"
    WAITING = "waiting"
    AWAITING_CALLBACK = "awaiting_callback"
    ABORTED = "aborted"
    FAILED = "failed"
    API_UNAVAILABLE = "api_unavailable"
    INCONSISTENT_DATA = "inconsistent_data"
    COMPLETED = "completed"
    RESUMED = "resumed"


class StepStatus(strEnum):
    SUCCESS = "success"
    SKIPPED = "skipped"
    SUSPEND = "suspend"
    WAITING = "waiting"
    AWAITING_CALLBACK = "awaiting_callback"
    FAILED = "failed"
    ABORT = "abort"
    COMPLETE = "complete"


class Process(Generic[S]):
    """ADT base class.

    This class defines an Algebraic Data Type - specifically a "sum type" - that defines the possible
    variants of a Process. It encapsulates the state and allows to fold _instances_ of a process into
    a single value. These instances correspond to subsequent steps of the process.
    """

    def __init__(self, s: S):
        self.s = s

    def map(self, f: Callable[[S], S]) -> Process[S]:
        """Apply a function to the process.

        >>> inc = lambda n: n + 1

        >>> Success(1).map(inc)
        Success 2

        >>> Skipped(1).map(inc)
        Skipped 2

        >>> Suspend(1).map(inc)
        Suspend 2

        >>> Waiting(1).map(inc)
        Waiting 2

        >>> AwaitingCallback(1).map(inc)
        AwaitingCallback 2

        >>> Abort(1).map(inc)
        Abort 2

        >>> Failed(1).map(inc)
        Failed 2

        >>> Complete(1).map(inc)
        Complete 2
        """

        def g(x: S) -> Process[S]:
            self_ = self.__class__
            return self_(f(x))

        return self._fold(g, g, g, g, g, g, g, g)

    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        """Unwrap the state from the Process category.

        >>> Success('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Success 'a'

        >>> Skipped('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Skipped 'a'

        >>> Suspend('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Suspend 'a'

        >>> Waiting('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Waiting 'a'

        >>> AwaitingCallback('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        AwaitingCallback 'a'

        >>> Abort('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Abort 'a'

        >>> Failed('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Failed 'a'

        >>> Complete('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Complete 'a'

        >>> Process('a')._fold(Success, Skipped, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)
        Traceback (most recent call last):
            ...
        NotImplementedError: Abstract function `_fold` must be implemented by the type constructor
        """
        raise NotImplementedError("Abstract function `_fold` must be implemented by the type constructor")

    def unwrap(self) -> S:
        """Get unwrapped state.

        >>> Success('a').unwrap()
        'a'

        >>> Skipped('a').unwrap()
        'a'

        >>> Suspend('a').unwrap()
        'a'

        >>> Waiting('a').unwrap()
        'a'

        >>> AwaitingCallback('a').unwrap()
        'a'

        >>> Abort('a').unwrap()
        'a'

        >>> Failed('a').unwrap()
        'a'

        >>> Complete('a').unwrap()
        'a'
        """
        return self._fold(identity, identity, identity, identity, identity, identity, identity, identity)

    def issuccess(self) -> bool:
        """Test if this instance is Success.

        >>> Success('a').issuccess()
        True

        >>> Skipped('a').issuccess()
        False

        >>> Suspend('a').issuccess()
        False

        >>> Waiting('a').issuccess()
        False

        >>> AwaitingCallback('a').issuccess()
        False

        >>> Abort('a').issuccess()
        False

        >>> Failed('a').issuccess()
        False

        >>> Complete('a').issuccess()
        False
        """
        return self._fold(
            const(True),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
        )

    def isskipped(self) -> bool:
        """Test if this instance is Skipped.

        >>> Success('a').isskipped()
        False

        >>> Skipped('a').isskipped()
        True

        >>> Suspend('a').isskipped()
        False

        >>> Waiting('a').isskipped()
        False

        >>> AwaitingCallback('a').isskipped()
        False

        >>> Abort('a').isskipped()
        False

        >>> Failed('a').isskipped()
        False

        >>> Complete('a').isskipped()
        False
        """
        return self._fold(
            const(False),
            const(True),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
        )

    def issuspend(self) -> bool:
        """Test if this instance is Suspend.

        >>> Success('a').issuspend()
        False

        >>> Skipped('a').issuspend()
        False

        >>> Suspend('a').issuspend()
        True

        >>> Waiting('a').issuspend()
        False

        >>> AwaitingCallback('a').issuspend()
        False

        >>> Abort('a').issuspend()
        False

        >>> Failed('a').issuspend()
        False

        >>> Complete('a').issuspend()
        False
        """
        return self._fold(
            const(False),
            const(False),
            const(True),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
        )

    def iswaiting(self) -> bool:
        """Test if this instance is Waiting.

        >>> Success('a').iswaiting()
        False

        >>> Skipped('a').iswaiting()
        False

        >>> Suspend('a').iswaiting()
        False

        >>> Waiting('a').iswaiting()
        True

        >>> AwaitingCallback('a').iswaiting()
        False

        >>> Abort('a').iswaiting()
        False

        >>> Failed('a').iswaiting()
        False

        >>> Complete('a').iswaiting()
        False
        """
        return self._fold(
            const(False),
            const(False),
            const(False),
            const(True),
            const(False),
            const(False),
            const(False),
            const(False),
        )

    def isawaitingcallback(self) -> bool:
        """Test if this instance is AwaitingCallback.

        >>> Success('a').isawaitingcallback()
        False

        >>> Skipped('a').isawaitingcallback()
        False

        >>> Suspend('a').isawaitingcallback()
        False

        >>> Waiting('a').isawaitingcallback()
        False

        >>> AwaitingCallback('a').isawaitingcallback()
        True

        >>> Abort('a').isawaitingcallback()
        False

        >>> Failed('a').isawaitingcallback()
        False

        >>> Complete('a').isawaitingcallback()
        False
        """
        return self._fold(
            const(False),
            const(False),
            const(False),
            const(False),
            const(True),
            const(False),
            const(False),
            const(False),
        )

    def isabort(self) -> bool:
        """Test if this instance is Abort.

        >>> Success('a').isabort()
        False

        >>> Skipped('a').isabort()
        False

        >>> Suspend('a').isabort()
        False

        >>> Waiting('a').isabort()
        False

        >>> AwaitingCallback('a').isabort()
        False

        >>> Abort('a').isabort()
        True

        >>> Failed('a').isabort()
        False

        >>> Complete('a').isabort()
        False
        """
        return self._fold(
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(True),
            const(False),
            const(False),
        )

    def isfailed(self) -> bool:
        """Test if this instance is Waiting.

        >>> Success('a').isfailed()
        False

        >>> Skipped('a').isfailed()
        False

        >>> Suspend('a').isfailed()
        False

        >>> Waiting('a').isfailed()
        False

        >>> AwaitingCallback('a').isfailed()
        False

        >>> Abort('a').isfailed()
        False

        >>> Failed('a').isfailed()
        True

        >>> Complete('a').isfailed()
        False
        """
        return self._fold(
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(True),
            const(False),
        )

    def iscomplete(self) -> bool:
        """Test if this instance is Complete.

        >>> Success('a').iscomplete()
        False

        >>> Skipped('a').iscomplete()
        False

        >>> Suspend('a').iscomplete()
        False

        >>> Waiting('a').iscomplete()
        False

        >>> AwaitingCallback('a').iscomplete()
        False

        >>> Abort('a').iscomplete()
        False

        >>> Failed('a').iscomplete()
        False

        >>> Complete('a').iscomplete()
        True
        """
        return self._fold(
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(False),
            const(True),
        )

    def __eq__(self, other: object) -> bool:
        """Test two instances for equality.

        >>> Success('a') == Success('a')
        True

        >>> Success('a') != Success('b')
        True

        >>> Success('a') != Suspend('a')
        True

        >>> Success('a') != Waiting('a')
        True

        >>> Success('a') != AwaitingCallback('a')
        True

        >>> Suspend('a') != Abort('a')
        True

        >>> Success('a') != 'a'
        True
        """
        return self.__class__ == other.__class__ and self.s == cast(Process, other).s

    @property
    def status(self) -> StepStatus:
        """Show status.

        >>> Success({}).status
        <StepStatus.SUCCESS: 'success'>

        >>> Skipped({}).status
        <StepStatus.SKIPPED: 'skipped'>

        >>> Suspend({}).status
        <StepStatus.SUSPEND: 'suspend'>

        >>> Waiting({}).status
        <StepStatus.WAITING: 'waiting'>

        >>> AwaitingCallback({}).status
        <StepStatus.AWAITING_CALLBACK: 'awaiting_callback'>

        >>> Abort({}).status
        <StepStatus.ABORT: 'abort'>

        >>> Failed({}).status
        <StepStatus.FAILED: 'failed'>

        >>> Complete({}).status
        <StepStatus.COMPLETE: 'complete'>
        """
        ss = getattr(self, "__name__", self.__class__.__name__).upper()
        return StepStatus[ss]

    @staticmethod
    def from_status(status: StepStatus, state: S) -> Process | None:
        """Make Process based on status and state.

        >>> Process.from_status('success', {})
        Success {}

        >>> Process.from_status('skipped', {})
        Skipped {}

        >>> Process.from_status('suspend', {})
        Suspend {}

        >>> Process.from_status('waiting', {})
        Waiting {}

        >>> Process.from_status('awaiting_callback', {})
        AwaitingCallback {}

        >>> Process.from_status('abort', {})
        Abort {}

        >>> Process.from_status('failed', {})
        Failed {}

        >>> Process.from_status('complete', {})
        Complete {}

        >>> Process.from_status('unknown', {})

        """
        status_class = _STATUSES.get(status)
        return status_class(state) if status_class else None

    @property
    def overall_status(self) -> ProcessStatus:
        """Show overall status of process or task.

        >>> Success({}).overall_status
        <ProcessStatus.RUNNING: 'running'>

        >>> Skipped({}).overall_status
        <ProcessStatus.RUNNING: 'running'>

        >>> Suspend({}).overall_status
        <ProcessStatus.SUSPENDED: 'suspended'>

        >>> Waiting({}).overall_status
        <ProcessStatus.WAITING: 'waiting'>

        >>> AwaitingCallback({}).overall_status
        <ProcessStatus.AWAITING_CALLBACK: 'awaiting_callback'>

        >>> Abort({}).overall_status
        <ProcessStatus.ABORTED: 'aborted'>

        >>> Failed({}).overall_status
        <ProcessStatus.FAILED: 'failed'>

        >>> Complete({}).overall_status
        <ProcessStatus.COMPLETED: 'completed'>
        """
        return self._fold(
            const(ProcessStatus.RUNNING),
            const(ProcessStatus.RUNNING),
            const(ProcessStatus.SUSPENDED),
            const(ProcessStatus.WAITING),
            const(ProcessStatus.AWAITING_CALLBACK),
            const(ProcessStatus.ABORTED),
            const(ProcessStatus.FAILED),
            const(ProcessStatus.COMPLETED),
        )

    def __repr__(self) -> str:
        """Show self.

        >>> repr(Success({}))
        'Success {}'

        >>> repr(Skipped({}))
        'Skipped {}'

        >>> repr(Suspend({}))
        'Suspend {}'

        >>> repr(Waiting({}))
        'Waiting {}'

        >>> repr(AwaitingCallback({}))
        'AwaitingCallback {}'

        >>> repr(Abort({}))
        'Abort {}'

        >>> repr(Failed({}))
        'Failed {}'

        >>> repr(Complete({}))
        'Complete {}'
        """
        name = self.__class__.__name__
        return f"{name} {self.s!r}"

    def on_success(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Success.

        >>> def assign_b(d):
        ...     d["b"] = 2
        ...     return d
        >>> Success({"a":1}).on_success(assign_b)
        Success {'a': 1, 'b': 2}

        >>> Failed({"a": 1}).on_success(assign_b)
        Failed {'a': 1}
        """
        return self.map(f) if self.issuccess() else self

    def on_skipped(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Skipped."""
        return self.map(f) if self.isskipped() else self

    def on_suspend(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Suspend."""
        return self.map(f) if self.issuspend() else self

    def on_waiting(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Waiting."""
        return self.map(f) if self.iswaiting() else self

    def on_awaiting_callback(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when AwaitingCallback."""
        return self.map(f) if self.isawaitingcallback() else self

    def on_abort(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Abort."""
        return self.map(f) if self.isabort() else self

    def on_failed(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Failed."""
        return self.map(f) if self.isfailed() else self

    def on_complete(self, f: Callable[[S], S]) -> Process[S]:
        """Apply function on Process state only when Complete."""
        return self.map(f) if self.iscomplete() else self

    def execute_step(self, step: Callable[[S], Process[S]]) -> Process[S]:
        """Execute a step transition based on a step function.

        A step can only be executed if the current state is success or skipped.

        >>> Success({"a":1}).execute_step(lambda s: Success(s))
        Success {'a': 1}
        >>> Success({"a":1}).execute_step(lambda s: Failed(s))
        Failed {'a': 1}

        >>> Waiting({"a":1}).execute_step(lambda s: Failed(s))
        Waiting {'a': 1}

        """

        return self._fold(step, step, Suspend, Waiting, AwaitingCallback, Abort, Failed, Complete)

    def abort(self) -> Process[S]:
        """Abort process.

        Always works except for completed processes
        """
        return self._fold(Abort, Abort, Abort, Abort, Abort, Abort, Abort, Complete)

    def resume(self, resume_suspend: Callable[[Process[S]], Process[S]]) -> Process[S]:
        """Resume process.

        Cannot resume Abort or Complete states

        Args:
            resume_suspend: function to call on resuming a suspended state. Might fail and determine the next state


        >>> Suspend({"a":1}).resume(lambda s: s)
        Success {'a': 1}

        >>> Success({"a":1}).resume(lambda s: Failed({"error": "Exception!!"}))
        Success {'a': 1}

        >>> Failed({"a":1}).resume(lambda s: Failed({"error": "Exception!!"}))
        Success {'a': 1}

        >>> Suspend({"a":1}).resume(lambda s: Failed({"error": "Exception!!"}))
        Failed {'error': 'Exception!!'}
        """

        next_state = self._fold(Success, Success, Success, Success, Success, Abort, Success, Complete)
        if self.issuspend() or self.isawaitingcallback():
            return resume_suspend(next_state)  # type: ignore

        return next_state  # type: ignore


class Success(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return success(self.s)


class Skipped(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return skipped(self.s)


class Suspend(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return suspend(self.s)


class Waiting(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return waiting(self.s)


class AwaitingCallback(Process[S]):
    __name__ = "Awaiting_Callback"

    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return awaiting_callback(self.s)


class Abort(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return abort(self.s)


class Failed(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return failed(self.s)


class Complete(Process[S]):
    def _fold(
        self,
        success: Callable[[S], F],
        skipped: Callable[[S], F],
        suspend: Callable[[S], F],
        waiting: Callable[[S], F],
        awaiting_callback: Callable[[S], F],
        abort: Callable[[S], F],
        failed: Callable[[S], F],
        complete: Callable[[S], F],
    ) -> F:
        return complete(self.s)


_STATUSES = {
    StepStatus.SUCCESS: Success,
    StepStatus.SKIPPED: Skipped,
    StepStatus.SUSPEND: Suspend,
    StepStatus.WAITING: Waiting,
    StepStatus.AWAITING_CALLBACK: AwaitingCallback,
    StepStatus.ABORT: Abort,
    StepStatus.FAILED: Failed,
    StepStatus.COMPLETE: Complete,
}

_NUM_STATUSES = len(_STATUSES)


def cond_bind(log: BoundLogger, state: dict[str, Any], key: str, as_key: str | None = None) -> BoundLogger:
    """Conditionally (on presence of key) build Structlog context."""
    if as_key is None:
        as_key = key
    if key in state:
        return log.bind(**{as_key: state[key]})
    return log


def log_mutations(old_process_state: State) -> Callable[[State], None]:
    def _log_mutations(new_process_state: State) -> None:
        mutations = {
            k: v for k, v in new_process_state.items() if k not in old_process_state or old_process_state[k] != v
        }
        logger.debug("Step returned a result state.", mutations=mutations)

    return _log_mutations


def errorlogger(error: ErrorDict) -> None:
    logger.error("Workflow returned an error.", **error)


def invalidate_status_counts() -> None:
    """Broadcast invalidate status counts to the websocket."""
    from orchestrator.websocket import broadcast_invalidate_status_counts

    broadcast_invalidate_status_counts()


def _exec_steps(steps: StepList, starting_process: Process, dblogstep: StepLogFuncInternal) -> Process:
    """Execute the workflow steps one by one until a Process state other than Success or Skipped is reached."""
    consolelogger = cond_bind(logger, starting_process.unwrap(), "reporter", "created_by")
    process = starting_process
    for step in steps:
        # Check if we need to continue with the process
        if not (process.issuccess() or process.isskipped()):
            break

        consolelogger = consolelogger.bind(step_name=step.name)

        # Debug logging of step information
        mutationlogger = log_mutations(process.unwrap())

        # Execute step
        try:
            engine_status = get_engine_settings()
            if engine_status.global_lock:
                # Exiting from thread workflow engine is Paused or Pausing
                consolelogger.info(
                    "Not executing Step as the workflow engine is Paused. Process will remain in state 'running'"
                )
                return process

            process = process.map(lambda s: s | {"__last_step_started_at": nowtz().timestamp()})
            step_result_process = process.execute_step(step)
        except Exception as e:
            consolelogger.error("An exception occurred while executing the workflow step.")
            step_result_process = Failed(e)

        # write the new process state after the step execution to the database
        # Convert ErrorState to ErrorDict when Failed or Waiting before writing to the database
        # as bare exceptions are not JSON serializable
        result_to_log = step_result_process.on_failed(error_state_to_dict).on_waiting(error_state_to_dict)
        result_to_log.on_success(mutationlogger).on_failed(errorlogger).on_waiting(errorlogger)
        process = dblogstep(step, result_to_log)
        # If database logging failed, the workflow should fail. When it was successful just continue with the
        # result of the executed step.
        consolelogger.debug("Workflow step executed.", process_status=process.status)

    return process


def runwf(pstat: ProcessStat, logstep: StepLogFunc) -> Process:
    """Run workflow optionally adding extra state.

    The extra state is used on resume and to set initial state
    """
    steps = pstat.log

    def _logstep(*x: Any) -> Process:
        return logstep(pstat, *x)

    logger.bind(workflow=pstat.workflow.name)

    def resume_suspend(process: Process) -> Process:
        state = process.unwrap()
        if "__step_group" in state:
            step = steps[0]
        else:
            step = steps.pop(0)
        return _logstep(step, process)

    next_state = pstat.state.resume(resume_suspend)
    # Set the step_log_fn in the contextvar.
    # This enables recursive step execution of sub steps with the same StepLogFunc.
    # Should probably be refactored at some point as contextvars is a kind of global state.
    step_log_fn_var.set(_logstep)
    executed_steps = _exec_steps(steps, next_state, _logstep)
    if executed_steps.overall_status == ProcessStatus.FAILED:
        invalidate_status_counts()
    return executed_steps


def abort_wf(pstat: ProcessStat, logstep: StepLogFunc) -> Process:
    """Abort a suspended workflow."""

    if not pstat.state.iscomplete():
        abort_func = make_step_function(Abort, "User Aborted")

        state = pstat.state.abort()

        return logstep(pstat, abort_func, state)
    return pstat.state


@_purestep("Start")
def init(state: State) -> Process:
    """Start of workflow."""
    return Success(state)


@_purestep("Done")
def done(state: State) -> Process:
    """End of workflow."""
    return Complete(state)


@_purestep("Abort")
def abort(state: State) -> Process:
    """End of aborted workflow."""
    return Abort(state)


# A `StepList` constructor to be used in reusable step sequences.
begin = StepList()



"""Set is_task=true on certain tasks.

This is required to make them appear in the completed tasks in the UI, and for the cleanup task to be able to
remove them.

Revision ID: 9736496e3eba
Revises: 961eddbd4c13
Create Date: 2025-12-10 16:42:29.060382

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9736496e3eba"
down_revision = "961eddbd4c13"
branch_labels = None
depends_on = None

task_names = [
    # Added in a76b9185b334
    "task_clean_up_tasks",
    "task_resume_workflows",
    # Added in 3c8b9185c221
    "task_validate_products",
    # Added in 961eddbd4c13
    "task_validate_subscriptions",
]


def upgrade() -> None:
    conn = op.get_bind()
    query = sa.text("UPDATE workflows SET is_task=true WHERE name = :task_name and is_task=false")
    for task_name in task_names:
        conn.execute(query, parameters={"task_name": task_name})


def downgrade() -> None:
    pass  # Does not make sense to downgrade back to a 'bad' state.



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Generic, TypeVar
from uuid import UUID

from sqlalchemy.orm import Query
from sqlalchemy.sql import Select

from orchestrator.db import (
    ProcessTable,
    ProductTable,
    SubscriptionTable,
    WorkflowTable,
)
from orchestrator.db.database import BaseModel
from orchestrator.search.core.types import EntityType, ExtractedField

from .traverse import (
    BaseTraverser,
    ProcessTraverser,
    ProductTraverser,
    SubscriptionTraverser,
    WorkflowTraverser,
)

ModelT = TypeVar("ModelT", bound=BaseModel)


@dataclass(frozen=True)
class EntityConfig(Generic[ModelT]):
    """A container for all configuration related to a specific entity type."""

    entity_kind: EntityType
    table: type[ModelT]

    traverser: "type[BaseTraverser]"
    pk_name: str
    root_name: str
    title_paths: list[str]  # List of field paths to check for title (with fallback)

    def get_all_query(self, entity_id: str | None = None) -> Query | Select:
        query = self.table.query
        if entity_id:
            pk_column = getattr(self.table, self.pk_name)
            query = query.filter(pk_column == UUID(entity_id))
        return query

    def get_title_from_fields(self, fields: list[ExtractedField]) -> str:
        """Extract title from fields using configured paths."""
        for title_path in self.title_paths:
            for field in fields:
                if field.path == title_path and field.value:
                    return str(field.value)
        return "UNKNOWN"


@dataclass(frozen=True)
class ProcessConfig(EntityConfig[ProcessTable]):
    """Processes need to eager load workflow for workflow_name field."""

    def get_all_query(self, entity_id: str | None = None) -> Query | Select:
        from sqlalchemy.orm import selectinload

        # Only load workflow, not subscriptions (keeps it lightweight)
        query = self.table.query.options(selectinload(ProcessTable.workflow))
        if entity_id:
            pk_column = getattr(self.table, self.pk_name)
            query = query.filter(pk_column == UUID(entity_id))
        return query


@dataclass(frozen=True)
class WorkflowConfig(EntityConfig[WorkflowTable]):
    """Workflows have a custom select() function that filters out deleted workflows."""

    def get_all_query(self, entity_id: str | None = None) -> Select:
        query = self.table.select()
        if entity_id:
            pk_column = getattr(self.table, self.pk_name)
            query = query.where(pk_column == UUID(entity_id))
        return query


ENTITY_CONFIG_REGISTRY: dict[EntityType, EntityConfig] = {
    EntityType.SUBSCRIPTION: EntityConfig(
        entity_kind=EntityType.SUBSCRIPTION,
        table=SubscriptionTable,
        traverser=SubscriptionTraverser,
        pk_name="subscription_id",
        root_name="subscription",
        title_paths=["subscription.description"],
    ),
    EntityType.PRODUCT: EntityConfig(
        entity_kind=EntityType.PRODUCT,
        table=ProductTable,
        traverser=ProductTraverser,
        pk_name="product_id",
        root_name="product",
        title_paths=["product.description", "product.name"],
    ),
    EntityType.PROCESS: ProcessConfig(
        entity_kind=EntityType.PROCESS,
        table=ProcessTable,
        traverser=ProcessTraverser,
        pk_name="process_id",
        root_name="process",
        title_paths=["process.workflow_name"],
    ),
    EntityType.WORKFLOW: WorkflowConfig(
        entity_kind=EntityType.WORKFLOW,
        table=WorkflowTable,
        traverser=WorkflowTraverser,
        pk_name="workflow_id",
        root_name="workflow",
        title_paths=["workflow.description", "workflow.name"],
    ),
}



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy import Select, Subquery, literal, select
from sqlalchemy_utils import Ltree

from orchestrator.db.models import AiSearchIndex
from orchestrator.search.core.types import SearchMetadata
from orchestrator.search.query.mixins import OrderDirection, StructuredOrderBy

from ..pagination import PageCursor
from .base import Retriever


def _apply_structured_ordering(stmt: Select, cand: Subquery, order_by: StructuredOrderBy | None = None) -> Select:
    if not order_by:
        return stmt.order_by(cand.c.entity_id.asc())

    path_subquery = (
        select(AiSearchIndex.value.label("order_value"))
        .where(AiSearchIndex.entity_id == cand.c.entity_id, AiSearchIndex.path == Ltree(order_by.element))
        .correlate(cand)
    ).scalar_subquery()

    _is_asc = order_by.direction == OrderDirection.ASC
    order_direction = path_subquery.asc() if _is_asc else path_subquery.desc()
    return stmt.order_by(order_direction)


class StructuredRetriever(Retriever):
    """Applies a dummy score for purely structured searches with no text query."""

    def __init__(self, cursor: PageCursor | None, order_by: StructuredOrderBy | None = None) -> None:
        self.cursor = cursor
        self.order_by = order_by

    def apply(self, candidate_query: Select) -> Select:
        cand = candidate_query.subquery()
        stmt = select(cand.c.entity_id, cand.c.entity_title, literal(1.0).label("score")).select_from(cand)

        if self.cursor is not None:
            stmt = stmt.where(cand.c.entity_id > self.cursor.id)

        return _apply_structured_ordering(stmt, cand, self.order_by)

    @property
    def metadata(self) -> SearchMetadata:
        return SearchMetadata.structured()



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pydantic_settings import BaseSettings

from orchestrator.utils.expose_settings import SettingsEnvVariablesSchema, SettingsExposedSchema

EXPOSED_ENV_SETTINGS_REGISTRY: dict[str, BaseSettings] = {}


def expose_settings(settings_name: str, base_settings: BaseSettings) -> BaseSettings:
    """Decorator to register settings classes."""
    EXPOSED_ENV_SETTINGS_REGISTRY[settings_name] = base_settings
    return base_settings


def get_all_exposed_settings() -> list[SettingsExposedSchema]:
    """Return all registered settings as dicts."""

    def _get_settings_env_variables(base_settings: BaseSettings) -> list[SettingsEnvVariablesSchema]:
        """Get environment variables from settings."""
        settings_env_variables: list[SettingsEnvVariablesSchema] = [
            SettingsEnvVariablesSchema(env_name=key, env_value=value)
            for key, value in base_settings.model_dump().items()
        ]
        return sorted(settings_env_variables, key=lambda v: v.env_name)

    return [
        SettingsExposedSchema(name=name, variables=_get_settings_env_variables(base_settings))
        for name, base_settings in EXPOSED_ENV_SETTINGS_REGISTRY.items()
    ]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from uuid import UUID

from fastapi.param_functions import Body
from fastapi.routing import APIRouter

from orchestrator.api.error_handling import raise_status
from orchestrator.db import db
from orchestrator.db.models import WorkflowTable
from orchestrator.schemas.workflow import WorkflowPatchSchema, WorkflowSchema

router = APIRouter()


@router.get("/{workflow_id}", response_model=WorkflowSchema)
def get_workflow_description(workflow_id: UUID) -> str:
    workflow = db.session.get(WorkflowTable, workflow_id)
    if workflow is None:
        raise_status(HTTPStatus.NOT_FOUND)
    return workflow


@router.patch("/{workflow_id}", status_code=HTTPStatus.CREATED, response_model=WorkflowSchema)
async def patch_workflow_by_id(workflow_id: UUID, data: WorkflowPatchSchema = Body(...)) -> WorkflowTable:
    workflow = db.session.get(WorkflowTable, workflow_id)
    if not workflow:
        raise_status(HTTPStatus.NOT_FOUND, f"Workflow id {workflow_id} not found")

    return await _patch_workflow_description(data, workflow)


async def _patch_workflow_description(
    data: WorkflowPatchSchema,
    workflow: WorkflowTable,
) -> WorkflowTable:

    updated_properties = data.model_dump(exclude_unset=True)
    description = updated_properties.get("description", workflow.description)
    workflow.description = description
    db.session.commit()
    return workflow



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handlers for search and aggregation execution with persistence."""

from uuid import UUID

import structlog

from orchestrator.db import SearchQueryTable
from orchestrator.db.database import WrappedSession
from orchestrator.search.query import engine
from orchestrator.search.query.queries import AggregateQuery, CountQuery, SelectQuery
from orchestrator.search.query.results import QueryResultsResponse, SearchResponse
from orchestrator.search.query.state import QueryState

logger = structlog.get_logger(__name__)


async def execute_search_with_persistence(
    query: SelectQuery,
    db_session: WrappedSession,
    run_id: UUID | None,
) -> tuple[SearchResponse, UUID, UUID]:
    """Execute search, persist to DB, return response and IDs.

    Args:
        query: SelectQuery for search operation
        db_session: Database session
        run_id: Existing run ID or None to create new one

    Returns:
        Tuple of (search_response, run_id, query_id)
    """
    # run_id must be provided when called from agent (endpoint creates the run)
    if run_id is None:
        raise ValueError("run_id is required - agent run must be created by endpoint with thread_id")

    # Execute search
    search_response = await engine.execute_search(query, db_session)

    # Save to database
    query_embedding = search_response.query_embedding
    query_state = QueryState(query=query, query_embedding=query_embedding)
    query_number = db_session.query(SearchQueryTable).filter_by(run_id=run_id).count() + 1
    search_query = SearchQueryTable.from_state(
        state=query_state,
        run_id=run_id,
        query_number=query_number,
    )
    db_session.add(search_query)
    db_session.commit()
    db_session.expire_all()

    logger.debug("Saved search query", query_id=str(search_query.query_id), query_number=query_number)
    return search_response, run_id, search_query.query_id


async def execute_aggregation_with_persistence(
    query: CountQuery | AggregateQuery,
    db_session: WrappedSession,
    run_id: UUID | None,
) -> tuple[QueryResultsResponse, UUID, UUID]:
    """Execute aggregation, persist to DB, return response and IDs.

    Args:
        query: CountQuery or AggregateQuery for aggregation operations
        db_session: Database session
        run_id: Existing run ID or None to create new one

    Returns:
        Tuple of (aggregation_response, run_id, query_id)
    """
    # run_id must be provided when called from agent (endpoint creates the run)
    if run_id is None:
        raise ValueError("run_id is required - agent run must be created by endpoint with thread_id")

    # Execute aggregation
    aggregation_response = await engine.execute_aggregation(query, db_session)

    # Save to database
    query_state = QueryState(query=query, query_embedding=None)
    query_number = db_session.query(SearchQueryTable).filter_by(run_id=run_id).count() + 1
    search_query = SearchQueryTable.from_state(
        state=query_state,
        run_id=run_id,
        query_number=query_number,
    )
    db_session.add(search_query)
    db_session.commit()
    db_session.expire_all()

    logger.debug("Saved aggregation query", query_id=str(search_query.query_id), query_number=query_number)

    return aggregation_response, run_id, search_query.query_id



from orchestrator.db.sorting.sorting import (
    QueryType,
    Sort,
    SortOrder,
    generic_apply_sorting,
    generic_column_sort,
    generic_sort,
    generic_sorts_validate,
)

__all__ = [
    "QueryType",
    "Sort",
    "SortOrder",
    "generic_sort",
    "generic_apply_sorting",
    "generic_column_sort",
    "generic_sorts_validate",
]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Literal

from pydantic import BaseModel, Field
from sqlalchemy import TEXT, bindparam
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy_utils.types.ltree import Ltree

from orchestrator.search.core.types import LTREE_SEPARATOR, FilterOp, SQLAColumn


class LtreeFilter(BaseModel):
    """Filter for ltree path operations."""

    op: Literal[
        FilterOp.MATCHES_LQUERY,
        FilterOp.IS_ANCESTOR,
        FilterOp.IS_DESCENDANT,
        FilterOp.PATH_MATCH,
        FilterOp.HAS_COMPONENT,
        FilterOp.NOT_HAS_COMPONENT,
        FilterOp.ENDS_WITH,
    ]
    value: str = Field(description="The ltree path or lquery pattern to compare against.")

    def to_expression(self, column: SQLAColumn, path: str) -> ColumnElement[bool]:
        """Converts the filter condition into a SQLAlchemy expression."""
        match self.op:
            case FilterOp.IS_DESCENDANT:
                ltree_value = Ltree(self.value)
                return column.op("<@")(ltree_value)
            case FilterOp.IS_ANCESTOR:
                ltree_value = Ltree(self.value)
                return column.op("@>")(ltree_value)
            case FilterOp.MATCHES_LQUERY:
                param = bindparam(None, self.value, type_=TEXT)
                return column.op("~")(param)
            case FilterOp.PATH_MATCH:
                ltree_value = Ltree(path)
                return column == ltree_value
            case FilterOp.HAS_COMPONENT | FilterOp.NOT_HAS_COMPONENT:
                return column.op("~")(bindparam(None, f"*{LTREE_SEPARATOR}{self.value}{LTREE_SEPARATOR}*", type_=TEXT))
            case FilterOp.ENDS_WITH:
                return column.op("~")(bindparam(None, f"*{LTREE_SEPARATOR}{self.value}", type_=TEXT))



import re
from collections.abc import Callable, Iterable, Iterator
from enum import Enum
from itertools import chain
from typing import Any, Union, cast

import structlog
from sqlalchemy import BinaryExpression, ColumnElement, CompoundSelect, Select, false, not_, or_
from sqlalchemy.orm import InstrumentedAttribute, MappedColumn

from orchestrator.db.database import BaseModel
from orchestrator.utils.helpers import camel_to_snake

logger = structlog.getLogger(__name__)


class Token(Enum):
    WORD = "WORD"
    OP_OR = "|"
    OP_NEG = "-"
    ASTERISK = "*"
    LPAREN = "("
    RPAREN = ")"
    COLON = ":"
    QUOTE = '"'
    END = "$"

    def __repr__(self) -> str:
        return self.name


SPECIAL_CHARS = {
    "|": Token.OP_OR,
    "-": Token.OP_NEG,
    "*": Token.ASTERISK,
    "(": Token.LPAREN,
    ")": Token.RPAREN,
    ":": Token.COLON,
    '"': Token.QUOTE,
}


class ParseError(RuntimeError):
    pass


Lexeme = Union[tuple[Token], tuple[Token, Any]]


class Lexer:
    word_boundary_regex = re.compile(r'[()|!:<>*\s"]')

    def __init__(self, source: str):
        self.source = source
        self.pos = 0

    def peek(self) -> str | None:
        return self.source[self.pos] if self.pos < len(self.source) else None

    def next_char(self) -> str | None:
        while self.pos < len(self.source):
            ch = self.source[self.pos]
            self.pos += 1
            return ch
        return None

    def read_word(self) -> str:
        chars = []
        while (ch := self.next_char()) and not re.fullmatch(Lexer.word_boundary_regex, ch):
            chars.append(ch)

        # Rewind one, since we read something not part of the word
        if ch:
            self.pos -= 1
        return "".join(chars)

    def lex(self) -> Iterable[Lexeme]:
        self.pos = 0
        max_pos = len(self.source)

        while self.pos < max_pos:
            ch = cast(str, self.peek())  # We know ch can't be None here
            if ch in SPECIAL_CHARS:
                self.pos += 1
                yield SPECIAL_CHARS[ch],
            elif re.fullmatch(Lexer.word_boundary_regex, ch):
                self.pos += 1
            else:
                word = self.read_word()
                yield Token.WORD, word

        yield Token.END,


Node = tuple[str, Any]


class Parser:
    """Class for creating a parse tree from the list of tokens.

    The following grammar is used. Nodenames starting with a _ will not appear in the final parse tree.
    <Query> ::= <AndExpression> ('|' <AndExpression>)*
    <AndExpression> ::= <_Term>*
    <_Term> ::= [<Negation>] <_PositiveTerm>
    <Negation> ::= '-'
    <_PositiveTerm> ::= <Group> | <_Value> | <KVTerm>
    <Group> ::= '(' <Query> ')'
    <Phrase> ::= '"' (<Word> | <PrefixWord>)* '"'
    <KVTerm> ::= <Value> ':' (<_Value> | <ValueGroup>)
    <_SearchWord> ::= <Word> | <PrefixWord>
    <PrefixWord> ::= <Word> '*'
    <Word> ::= <text without "*()!:<> or whitespace>
    <_Value> ::= <Phrase> | <_SearchWord>
    <ValueGroup> ::= '(' <Value> ('|' <Value>)* ')'
    """

    def __init__(self, tokens: Iterable[Lexeme]):
        """Args: tokens - An iterable yielding lexemes.

        `tokens` must not be empty. The result from the Lexer class
        will always end with (END,) and so is never empty.
        Ensure that next_token() is never called after the (END,)
        token is consumed.
        """
        self.tokens = tokens
        self.it: Iterator[Lexeme] = iter([])

    def peek(self) -> Lexeme:
        token = next(self.it)
        self.it = chain([token], self.it)
        return token

    def next_token(self) -> Lexeme:
        return next(self.it)

    def error(self, msg: str) -> None:
        logger.debug(msg)

    def expect(self, expected_type: Token) -> Lexeme:
        nt = self.next_token()
        if nt[0] != expected_type:
            msg = f"Expected {expected_type}, got {nt[0]}"
            self.error(msg)
            raise ParseError(msg)
        return nt

    def parse_value_group(self) -> Node | None:
        self.expect(Token.LPAREN)
        value = self.parse_value()
        if not value:
            return None
        values = [value]
        while next_token := self.peek():
            if next_token[0] == Token.RPAREN:
                break
            if next_token[0] == Token.OP_OR:
                self.expect(Token.OP_OR)
                if next_value := self.parse_value():
                    values.append(next_value)
            else:
                raise ParseError(f"Expected a Value or ')', got {next_token}")

        self.expect(Token.RPAREN)
        return "ValueGroup", values

    def parse_value(self) -> Node | None:
        next_token = self.peek()
        if next_token[0] == Token.QUOTE:
            return self.parse_phrase()
        if next_token[0] == Token.WORD:
            return self.parse_search_word()
        return None

    def parse_search_word(self) -> Node:
        word = cast(tuple, self.expect(Token.WORD))
        next_token = self.peek()
        if next_token[0] == Token.ASTERISK:
            self.next_token()
            return "PrefixWord", word[1]
        return "Word", word[1]

    def parse_phrase(self) -> Node:
        self.expect(Token.QUOTE)
        search_words = []
        while not self.peek()[0] == Token.QUOTE:
            if not (self.peek()[0] == Token.WORD):
                self.error(f"Expected a Word or PrefixWord inside phrase. Got: {self.peek()}")
                break

            search_words.append(self.parse_search_word())

        self.expect(Token.QUOTE)
        return "Phrase", search_words

    def parse_group(self) -> Node:
        self.expect(Token.LPAREN)
        query = self.parse_query()
        self.expect(Token.RPAREN)
        return "Group", query[1] if query else []

    def parse_positive_term(self) -> Node | None:
        # Should return None if not a valid term
        token = self.peek()
        if token[0] == Token.LPAREN:
            return self.parse_group()

        value = self.parse_value()
        if not value:
            return None
        next_token = self.peek()
        if next_token and next_token[0] == Token.COLON:
            # It's a KVTerm, e.g. status:active
            self.next_token()
            next_token = self.peek()
            kv_value = self.parse_value_group() if next_token[0] == Token.LPAREN else self.parse_value()

            # An empty value is normally not a problem, but it is in a KVTerm, so we check for it.
            if not kv_value or kv_value[0] == "Phrase" and not kv_value[1]:
                raise ParseError("Value term in KVTerm can not be empty.")
            return "KVTerm", (value, kv_value)
        return value

    def parse_term(self) -> Node | None:
        token = self.peek()
        if token[0] == Token.OP_NEG:
            self.next_token()
            term = self.parse_positive_term()
            return ("Negation", term) if term else None
        return self.parse_positive_term()

    def parse_and_expression(self) -> Node | None:
        terms = []
        while term := self.parse_term():
            terms.append(term)

        return ("AndExpression", terms) if terms else None

    def parse_query(self) -> Node:
        expressions = []
        while True:
            and_expression = self.parse_and_expression()
            if and_expression:
                expressions.append(and_expression)
            if self.peek()[0] not in [Token.END, Token.RPAREN]:
                self.expect(Token.OP_OR)
            else:
                break
        return "Query", expressions

    def parse(self) -> Node:
        self.it = iter(self.tokens)
        query = self.parse_query()
        self.expect(Token.END)
        return query


class TSQueryVisitor:
    _glue_chars = re.compile(r"[-_@#$%^&]")

    @staticmethod
    def _split_term(term: str) -> list[str]:
        return [part.strip() for part in TSQueryVisitor._glue_chars.split(term) if part and not part.isspace()]

    @staticmethod
    def visit_group(node: Node, acc: list[str]) -> None:
        acc.append("(")
        for expr in node[1]:
            TSQueryVisitor.visit_and_expression(expr, acc)
            acc.append(" | ")
        acc.pop()
        acc.append(")")

    @staticmethod
    def visit_kv_term(node: Node, acc: list[str]) -> None:
        key_node, value_node = node[1]

        # Re-use visit_term
        if key_node[0] == "Word":
            # Camel case key term
            key_node = ["Word", camel_to_snake(key_node[1])]
        TSQueryVisitor.visit_term(key_node, acc)
        acc.append(" <-> ")

        if value_node[0] == "ValueGroup":
            acc.append("(")
            for v in value_node[1]:
                TSQueryVisitor.visit_term(v, acc)
                acc.append(" | ")
            acc.pop()
            acc.append(")")
        else:
            TSQueryVisitor.visit_term(value_node, acc)

    @staticmethod
    def visit_search_word(node: Node, acc: list[str]) -> None:

        if node[0] not in ["Word", "PrefixWord"]:
            raise Exception(f"Invalid SearchWord type {node[0]}")

        # Postgres is finicky with single quotes in the query. In some situations it throws an error.
        # To avoid the special handling of ' by PG, we replace it with double quotes.

        text = node[1].replace("'", '"')

        parts = TSQueryVisitor._split_term(text)

        if not parts:
            return

        text = " <-> ".join(parts)

        if node[0] == "PrefixWord":
            text = f"{text}:*"

        # Wrap parentheses for clarity, even though <-> has the highest operator precedence
        text = f"({text})" if len(parts) > 1 else text
        acc.append(text)

    @staticmethod
    def visit_phrase(node: Node, acc: list[str]) -> None:
        for search_word in node[1]:
            TSQueryVisitor.visit_search_word(search_word, acc)
            acc.append(" <-> ")

        if len(node[1]):
            acc.pop()  # Remove trailing space

    @staticmethod
    def visit_term(node: Node, acc: list[str]) -> None:
        node_type = node[0]
        if node_type == "Phrase":
            TSQueryVisitor.visit_phrase(node, acc)
        elif node_type in ["Word", "PrefixWord"]:
            TSQueryVisitor.visit_search_word(node, acc)
        elif node_type == "KVTerm":
            TSQueryVisitor.visit_kv_term(node, acc)
        elif node_type == "Group":
            TSQueryVisitor.visit_group(node, acc)
        elif node_type == "Negation":
            acc.append("!")
            should_group = node[1][0] in ["Phrase", "KVTerm"]
            if should_group:
                acc.append("(")
            TSQueryVisitor.visit_term(node[1], acc)
            if should_group:
                acc.append(")")
        else:
            raise Exception(f"Unexpected term node type: {node_type}")

    @staticmethod
    def visit_and_expression(node: Node, acc: list[str]) -> None:
        for term in node[1]:
            TSQueryVisitor.visit_term(term, acc)
            acc.append(" & ")
        if len(node[1]):
            acc.pop()

    @staticmethod
    def visit_query(node: Node, acc: list[str]) -> None:
        for expression in node[1]:
            TSQueryVisitor.visit_and_expression(expression, acc)
            acc.append(" | ")

        if len(node[1]):
            acc.pop()

    @staticmethod
    def visit(parse_tree: Node) -> str:
        acc: list[str] = []
        node_type = parse_tree[0]
        if node_type == "Query":
            TSQueryVisitor.visit_query(parse_tree, acc)
        return "".join(acc)


WhereCondGenerator = Callable[[Node], Union[BinaryExpression, ColumnElement[bool]]]


class SQLAlchemyVisitor:
    def __init__(
        self,
        stmt: Select,
        mappings: dict[str, WhereCondGenerator],
        *,
        base_table: type[BaseModel] | None = None,
        join_key: MappedColumn | InstrumentedAttribute | None = None,
    ):
        self.base_stmt = stmt
        self.mappings = mappings
        # base_table and join_key are required for group terms, because of the union join
        # If the query does not use grouping, these parameters are optional
        self.base_table = base_table
        self.join_key = join_key

    def visit_kv_term(self, stmt: Select, node: Node, is_negated: bool) -> Select:
        key_node, value_node = node[1]
        if key_node[0] == "Word":
            cond_expr_fn = self.mappings.get(camel_to_snake(key_node[1]))
            if not cond_expr_fn:
                # Non-existing columns will emit `where false`, returning no results
                return stmt.where(false())
            if value_node[0] in ["Word", "PrefixWord", "Phrase"]:
                cond_expr = cond_expr_fn(value_node)
                return stmt.where(cond_expr if not is_negated else not_(cond_expr))
            if value_node[0] == "ValueGroup":
                ors = [cond_expr_fn(vg_node) for vg_node in value_node[1]]
                or_expr = or_(False, *ors)
                cond_expr2 = or_expr if not is_negated else not_(or_expr)
                return stmt.where(cond_expr2)

        # Only Word or single-term Phrase key-nodes are supported.
        if key_node[0] == "Phrase" and len(key_node[1]) == 1:
            return self.visit_kv_term(stmt, ("KVTerm", (key_node[1][0], value_node)), is_negated)
        return stmt

    def visit_search_word(self, stmt: Select, node: Node, is_negated: bool) -> Select:
        # The dynamically generated list of expression may not be empty, so we prepend with a "default" False
        or_expr = or_(False, *(cond_expr_fn(node) for cond_expr_fn in self.mappings.values()))
        return stmt.where(or_expr if not is_negated else not_(or_expr))

    def visit_group(self, stmt: Select, node: Node, is_negated: bool) -> Select:
        if self.join_key is None or self.base_table is None:
            logger.error("Group statements not supported without a base_table and a join_key")
            return stmt
        subquery = self.visit_query(self.base_stmt, node).cte()
        if is_negated:
            return stmt.outerjoin_from(
                self.base_table, subquery, self.join_key == subquery.c[self.join_key.name]
            ).where(subquery.c[self.join_key.name].is_(None))
        return stmt.join_from(self.base_table, subquery, self.join_key == subquery.c[self.join_key.name])

    def visit_term(self, stmt: Select, node: Node, is_negated: bool = False) -> Select:
        node_type = node[0]
        if node_type == "KVTerm":
            return self.visit_kv_term(stmt, node, is_negated)
        if node_type == "Negation":
            return self.visit_term(stmt, node[1], is_negated=True)
        if node_type in ["Word", "PrefixWord", "Phrase"]:
            return self.visit_search_word(stmt, node, is_negated)
        if node_type == "Group":
            return self.visit_group(stmt, node, is_negated)
        raise Exception(f"Unexpected term node type: {node_type}")

    def visit_and_expression(self, stmt: Select, node: Node) -> Select:
        for term in node[1]:
            stmt = self.visit_term(stmt, term)
        return stmt

    def visit_query(self, stmt: Select, node: Node) -> Select | CompoundSelect:
        stmt = self.visit_and_expression(stmt, node[1][0])
        if len(node[1]) > 1:
            return stmt.union(*(self.visit_and_expression(self.base_stmt, expression) for expression in node[1][1:]))
        return stmt

    def visit(self, parse_tree: Node) -> Select | CompoundSelect:
        node_type = parse_tree[0]
        if node_type == "Query":
            return self.visit_query(self.base_stmt, parse_tree)
        return self.base_stmt


def create_ts_query_string(search_query: str) -> str:
    return TSQueryVisitor.visit(Parser(Lexer(search_query).lex()).parse())


def create_sqlalchemy_select(
    stmt: Select,
    search_query: str,
    mappings: dict[str, WhereCondGenerator],
    base_table: type[BaseModel],
    join_key: MappedColumn | InstrumentedAttribute,
) -> Select | CompoundSelect:
    return SQLAlchemyVisitor(stmt, mappings, base_table=base_table, join_key=join_key).visit(
        Parser(Lexer(search_query).lex()).parse()
    )



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, AsyncIterator, cast

import structlog
from pydantic_ai import Agent
from pydantic_ai.ag_ui import StateDeps

from orchestrator.search.agent.events import (
    PlanCreatedTaskValue,
    RunContext,
    make_plan_created_event,
    make_step_active_event,
)
from orchestrator.search.agent.memory import MemoryScope, ToolStep
from orchestrator.search.agent.prompts import get_planning_prompt
from orchestrator.search.agent.skill_runner import SkillRunner
from orchestrator.search.agent.state import ExecutionPlan, SearchState, Task, TaskAction, TaskStatus
from orchestrator.search.agent.utils import log_agent_request, log_execution_plan

if TYPE_CHECKING:
    from orchestrator.search.agent.skills import Skill

logger = structlog.get_logger(__name__)


@dataclass
class Planner:
    """Creates and executes plans via LLM.

    Owns the full plan lifecycle: create plan, iterate tasks,
    run SkillRunners, handle replanning on failure.
    """

    model: str
    skills: dict[TaskAction, Skill]
    debug: bool = False

    def __post_init__(self) -> None:
        self._agent = Agent(
            model=self.model,
            deps_type=StateDeps[SearchState],
            output_type=ExecutionPlan,
            name="create_execution_plan",
            retries=2,
        )

    async def _create_plan(self, ctx: RunContext) -> ExecutionPlan:
        """Create an execution plan via LLM."""
        ctx.state.memory.start_step("Planner")

        logger.info("Planner: Creating execution plan")

        message_history = ctx.state.memory.get_message_history(max_turns=5, scope=MemoryScope.FULL)
        prompt = get_planning_prompt(ctx.state)

        if self.debug:
            log_agent_request("Planner", prompt, message_history)

        result = await self._agent.run(instructions=prompt, message_history=message_history, deps=StateDeps(ctx.state))

        plan = cast(ExecutionPlan, result.output)

        if self.debug:
            log_execution_plan(plan)

        logger.info(
            "Planner: Plan created",
            num_tasks=len(plan.tasks),
            tasks=[f"{i+1}. {t.reasoning}" for i, t in enumerate(plan.tasks)],
        )

        return plan

    async def _run_tasks(self, ctx: RunContext, tasks: list[Task]) -> AsyncIterator[Any]:
        """Execute tasks sequentially, yielding events. Sets task.status on each."""
        for task in tasks:
            skill = self.skills.get(task.action_type)
            if not skill:
                logger.warning(f"Unknown task type: {task.action_type}, skipping")
                continue

            if task.action_type in (TaskAction.SEARCH, TaskAction.AGGREGATION):
                ctx.state.query = None
                ctx.state.pending_filters = None

            task.status = TaskStatus.EXECUTING
            runner = SkillRunner(skill=skill, model=self.model, debug=self.debug)

            try:
                async for event in runner.run(ctx, reasoning=task.reasoning):
                    yield event
                task.status = TaskStatus.COMPLETED
            except Exception as e:
                logger.error(f"Task failed: {task.action_type}", error=str(e))
                task.status = TaskStatus.FAILED
                ctx.state.memory.record_tool_step(
                    ToolStep(
                        step_type="error", description=f"{skill.name} failed: {e}", success=False, error_message=str(e)
                    )
                )
                break

    async def execute(self, ctx: RunContext, *, target_action: TaskAction | None = None) -> AsyncIterator[Any]:
        """Create and execute a plan, streaming all events.

        Args:
            ctx: Current run context
            target_action: If set, skip planning and execute this single action directly
        """
        if target_action:
            tasks = [Task(action_type=target_action, reasoning="Direct invocation")]
        else:
            yield make_step_active_event("Planner")
            tasks = (await self._create_plan(ctx)).tasks

            # Yield full plan so frontend can render all steps as pending
            plan_tasks = [
                PlanCreatedTaskValue(skill_name=skill.name, reasoning=task.reasoning)
                for task in tasks
                if (skill := self.skills.get(task.action_type))
            ]
            yield make_plan_created_event(plan_tasks)

        async for event in self._run_tasks(ctx, tasks):
            yield event



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Search agent tools package."""

from orchestrator.search.agent.tools.aggregation import (
    aggregation_execution_toolset,
    aggregation_toolset,
    run_aggregation,
    set_aggregations,
    set_grouping,
    set_temporal_grouping,
)
from orchestrator.search.agent.tools.filters import (
    discover_filter_paths,
    ensure_query_initialized,
    filter_building_toolset,
    get_valid_operators,
    set_filter_tree,
)
from orchestrator.search.agent.tools.result_actions import (
    fetch_entity_details,
    prepare_export,
    result_actions_toolset,
)
from orchestrator.search.agent.tools.search import (
    run_search,
    search_execution_toolset,
)

__all__ = [
    "ensure_query_initialized",
    "aggregation_execution_toolset",
    "aggregation_toolset",
    "discover_filter_paths",
    "fetch_entity_details",
    "filter_building_toolset",
    "get_valid_operators",
    "prepare_export",
    "result_actions_toolset",
    "run_aggregation",
    "run_search",
    "search_execution_toolset",
    "set_aggregations",
    "set_filter_tree",
    "set_grouping",
    "set_temporal_grouping",
]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.settings import get_authorizers
from orchestrator.workflow import StepList, workflow

authorizers = get_authorizers()


# This workflow has been made to create the initial import process for a SN7 subscription
# it does not do anything but is needed for the correct showing in the GUI.
@workflow(
    "Dummy workflow to replace removed workflows",
    authorize_callback=authorizers.authorize_callback,
    retry_auth_callback=authorizers.retry_auth_callback,
)
def removed_workflow() -> StepList:
    return StepList()



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import strawberry
from pydantic import ConfigDict

from orchestrator.schemas.base import OrchestratorBaseModel
from pydantic_forms.types import strEnum


@strawberry.enum
class GlobalStatusEnum(strEnum):
    RUNNING = "RUNNING"
    PAUSED = "PAUSED"
    PAUSING = "PAUSING"


class EngineSettingsBaseSchema(OrchestratorBaseModel):
    global_lock: bool


class WorkerStatus(OrchestratorBaseModel):
    executor_type: str
    number_of_workers_online: int = 0
    number_of_queued_jobs: int = 0
    number_of_running_jobs: int = 0


class EngineSettingsSchema(EngineSettingsBaseSchema):
    global_status: GlobalStatusEnum | None = None
    running_processes: int
    model_config = ConfigDict(from_attributes=True)



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy

import structlog
from pydantic import ValidationError

from orchestrator import llm_settings
from orchestrator.db import db
from orchestrator.db.models import ProcessSubscriptionTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.settings import reset_search_index
from orchestrator.services.subscriptions import get_subscription
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle
from orchestrator.utils.json import to_serializable
from orchestrator.websocket import sync_invalidate_subscription_cache
from orchestrator.workflow import Step, step
from pydantic_forms.types import State, UUIDstr

logger = structlog.get_logger(__name__)


@step("Unlock subscription")
def resync(subscription: SubscriptionModel) -> State:
    """Transition a subscription to in sync."""
    subscription.insync = True
    sync_invalidate_subscription_cache(subscription.subscription_id)
    return {"subscription": subscription}


@step("Lock subscription")
def unsync(subscription_id: UUIDstr, __old_subscriptions__: dict | None = None) -> State:
    """Transition a subscription to out of sync.

    This step will also create a backup of the current subscription details in the state with the key
    `__old_subscriptions__`

    Note:  This step will NOT overwrite any existing data in the state for the `__old_subscriptions__` key. Some
    workflows already change subscription details in the initial form step. This is not a best practice but
    it can be handy/needed for some scenarios. To ensure that you get a correct backup these forms need to create
    the backup in the initial form step and save it to `__old_subscriptions__`.

    An example, showing the creation of a backup for 2 subscriptions during the initial form step :

    ```
    subscription = YOUR_DOMAIN_MODEL.from_subscription(subscription_id)
    subscription_backup = {}
    subscription_backup[subscription.subscription_id] = deepcopy(subscription.model_dump())
    subscription_backup[second_subscription.subscription_id] = deepcopy(second_subscription.model_dump())
    return {..., "__old_subscriptions__": subscription_backup }
    ```

    It's also possible to make the backup for a subscription that doesn't have a domain model:

    ```
    from orchestrator.utils.json import to_serializable
    subscription = subscriptions.get_subscription(subscription_id)
    subscription_backup = {subscription_id: to_serializable(subscription)}
    Do your changes here ...
    return {..., "subscription": to_serializable(subscription), "__old_subscriptions__": subscription_backup }
    ```

    """
    try:
        subscription = SubscriptionModel.from_subscription(subscription_id)
    except ValidationError:
        subscription = get_subscription(subscription_id)  # type: ignore

    # Handle backup if needed
    if __old_subscriptions__ and __old_subscriptions__.get(subscription_id):
        logger.info(
            "Skipping backup of subscription details because it already exists in the state",
            subscription_id=str(subscription_id),
        )
        subscription_backup = __old_subscriptions__
    else:
        logger.info("Creating backup of subscription details in the state", subscription_id=str(subscription_id))
        subscription_backup = __old_subscriptions__ or {}
        if isinstance(subscription, SubscriptionModel):
            subscription_backup[str(subscription_id)] = deepcopy(subscription.model_dump())
        else:
            subscription_backup[str(subscription_id)] = to_serializable(subscription)  # type: ignore

    # Handle transition
    if not subscription.insync:
        raise ValueError("Subscription is already out of sync, cannot continue!")
    subscription.insync = False
    sync_invalidate_subscription_cache(subscription.subscription_id)

    return {"subscription": subscription, "__old_subscriptions__": subscription_backup}


@step("Lock subscription")
def unsync_unchecked(subscription_id: UUIDstr) -> State:
    """Use for validation workflows that need to run if the subscription is out of sync."""
    subscription = SubscriptionModel.from_subscription(subscription_id)
    subscription.insync = False
    return {"subscription": subscription}


def store_process_subscription_relationship(process_id: UUIDstr, subscription_id: UUIDstr) -> ProcessSubscriptionTable:
    process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=subscription_id)
    db.session.add(process_subscription)
    return process_subscription


def store_process_subscription(workflow_target: Target | None = None) -> Step:
    if workflow_target:
        deprecation_warning = (
            "Providing a workflow target to function store_process_subscription() is deprecated. "
            "This information is already stored in the workflow table."
        )
        logger.warning(deprecation_warning)

    @step("Create Process Subscription relation")
    def _store_process_subscription(process_id: UUIDstr, subscription_id: UUIDstr) -> None:
        store_process_subscription_relationship(process_id, subscription_id)

    return _store_process_subscription


def set_status(status: SubscriptionLifecycle) -> Step:
    @step(f"Set subscription to '{status}'")
    def _set_status(subscription: SubscriptionModel) -> State:
        """Set subscription to status."""
        subscription = SubscriptionModel.from_other_lifecycle(subscription, status)
        return {"subscription": subscription}

    _set_status.__doc__ = f"Set subscription to '{status}'."
    return _set_status


@step("Refresh subscription search index")
def refresh_subscription_search_index(subscription: SubscriptionModel | None) -> State:
    """Refresh subscription search index.

    Args:
        subscription: Subscription to refresh search index.

    Returns:
        State of the workflow.

    """
    try:
        reset_search_index()
        if llm_settings.SEARCH_ENABLED and subscription:
            from orchestrator.search.core.types import EntityType
            from orchestrator.search.indexing import run_indexing_for_entity

            run_indexing_for_entity(EntityType.SUBSCRIPTION, str(subscription.subscription_id))
    except Exception:
        # Don't fail workflow in case of unexpected error
        logger.warning("Error updated the subscriptions search index")
    return {}


@step("Refresh process search index")
def refresh_process_search_index(process_id: UUIDstr | None) -> State:
    """Refresh process search index.

    Args:
        process_id: Process to refresh search index.

    Returns:
        State of the workflow.

    """
    try:
        reset_search_index()
        if llm_settings.SEARCH_ENABLED and process_id:
            from orchestrator.search.core.types import EntityType
            from orchestrator.search.indexing import run_indexing_for_entity

            run_indexing_for_entity(EntityType.PROCESS, process_id)
    except Exception:
        # Don't fail workflow in case of unexpected error
        logger.warning("Error updating the processes search index")
    return {}



from structlog import get_logger

from orchestrator import __version__
from orchestrator.graphql.resolvers.helpers import make_async
from orchestrator.graphql.schemas.version import VersionType
from orchestrator.graphql.types import OrchestratorInfo
from orchestrator.graphql.utils import create_resolver_error_handler

logger = get_logger(__name__)


VERSIONS = [f"orchestrator-core: {__version__}"]


@make_async
def resolve_version(info: OrchestratorInfo) -> VersionType | None:
    logger.debug("resolve_version() called")
    _error_handler = create_resolver_error_handler(info)

    ver = None
    try:
        ver = VersionType(application_versions=VERSIONS)
    except Exception as e:
        logger.error(f"Error getting version: {str(e)}")
        _error_handler("Failed to retrieve orchestrator_core version", extensions={"code": "PACKAGE_VERSION_ERROR"})

    return ver



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MCP (Model Context Protocol) adapter for the search agent.

MCPWorker owns the agent lifecycle and stream consumption — parallel to
AGUIWorker and A2AWorker.  It consumes ``agent.run_stream_events()`` to
collect ToolArtifact results and the final LLM output, then returns a
combined result string.

The FastMCP tool handlers are thin wrappers that delegate to the worker.

Unlike AG-UI (streaming SSE to a frontend) or A2A (task lifecycle with
broker/storage), the MCP adapter is **stateless** — each tool call is
independent with no session or memory between calls.
"""

from __future__ import annotations

import uuid
from contextlib import AsyncExitStack

import structlog
from mcp.server.fastmcp import FastMCP
from pydantic_ai.ag_ui import StateDeps

from orchestrator.db import db
from orchestrator.db.models import AgentRunTable
from orchestrator.search.agent.adapters.stream import collect_stream_output
from orchestrator.search.agent.agent import AgentAdapter
from orchestrator.search.agent.state import SearchState, TaskAction
from orchestrator.search.core.types import EntityType

logger = structlog.get_logger(__name__)


class MCPWorker:
    """Orchestrates MCP tool execution: state setup, stream consumption, result assembly.

    Parallel to AGUIWorker and A2AWorker — owns the agent reference and the
    full tool-call lifecycle so the FastMCP handlers stay thin.
    """

    def __init__(self, agent: AgentAdapter) -> None:
        self.agent = agent

    async def run_skill(self, query: str, target_action: TaskAction | None = None) -> str:
        """Create state, consume the agent stream, and return a result string."""
        deps = StateDeps(SearchState(user_input=query))

        deps.state.run_id = uuid.uuid4()
        agent_run = AgentRunTable(run_id=deps.state.run_id, thread_id=str(uuid.uuid4()), agent_type="mcp")
        db.session.add(agent_run)
        db.session.commit()

        logger.debug("MCPWorker: Starting skill execution", target_action=target_action, query=query[:100])

        try:
            event_stream = self.agent.run_stream_events(deps=deps, target_action=target_action)
            return await collect_stream_output(event_stream)
        except Exception:
            logger.exception("MCPWorker: Skill execution failed", query=query[:100])
            raise


class MCPApp:
    """MCP adapter app: FastMCP server, worker, tools, and lifecycle.

    Bundles the Starlette sub-app with the session manager lifecycle so the
    host can ``async with mcp_app`` symmetrically with A2AApp.  Tool handlers
    close over ``self.worker`` — no module-level global needed.
    """

    def __init__(self, agent: AgentAdapter) -> None:
        self.agent = agent
        self.worker = MCPWorker(agent)
        self._stack: AsyncExitStack

        self.server = FastMCP(
            name="WFO Search Agent",
            instructions="Search, filter and aggregate orchestration data",
            stateless_http=True,
        )
        self._register_tools()
        self.app = self.server.streamable_http_app()

    def _register_tools(self) -> None:
        worker = self.worker

        @self.server.tool()
        async def search(query: str) -> str:
            """Find subscriptions, products, workflows, or processes.

            Describe what you're looking for in natural language.
            Examples: "active subscriptions", "failed workflows from last week"
            """
            return await worker.run_skill(query, TaskAction.SEARCH)

        @self.server.tool()
        async def aggregate(query: str) -> str:
            """Count, sum, or average data with grouping.

            Describe what aggregation you need.
            Examples: "count subscriptions by product", "average workflow duration by status"
            """
            return await worker.run_skill(query, TaskAction.AGGREGATION)

        @self.server.tool()
        async def get_entity_details(entity_type: EntityType, entity_id: uuid.UUID) -> str:
            """Fetch full details for a specific entity.

            Args:
                entity_type: The type of entity (SUBSCRIPTION, PRODUCT, WORKFLOW, PROCESS)
                entity_id: The UUID of the entity
            """
            query = f"Get details for {entity_type.value} {entity_id}"
            return await worker.run_skill(query, TaskAction.RESULT_ACTIONS)

        @self.server.tool()
        async def ask(query: str) -> str:
            """Ask any question about the orchestration system.

            The agent will determine the best approach — search, aggregate, or answer directly.
            """
            return await worker.run_skill(query, target_action=None)

    async def __aenter__(self) -> MCPApp:
        self._stack = AsyncExitStack()
        await self._stack.__aenter__()
        await self._stack.enter_async_context(self.server.session_manager.run())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore[no-untyped-def]
        await self._stack.__aexit__(exc_type, exc_val, exc_tb)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from dataclasses import dataclass

from ag_ui.core import CustomEvent
from pydantic import Field
from typing_extensions import TypedDict

from orchestrator.search.agent.state import SearchState
from orchestrator.search.agent.utils import current_timestamp_ms


class AgentStepActiveValue(TypedDict, total=False):
    """Typed structure for AGENT_STEP_ACTIVE event value."""

    step: str
    reasoning: str


class AgentStepActiveEvent(CustomEvent):
    """Event emitted when an agent step becomes active."""

    name: str = Field(default="AGENT_STEP_ACTIVE", frozen=True)
    value: AgentStepActiveValue


def make_step_active_event(step_name: str, reasoning: str | None = None) -> AgentStepActiveEvent:
    """Create an AGENT_STEP_ACTIVE event for yielding into the event stream."""
    value = AgentStepActiveValue(step=step_name)
    if reasoning:
        value["reasoning"] = reasoning
    return AgentStepActiveEvent(timestamp=current_timestamp_ms(), value=value)


class PlanCreatedTaskValue(TypedDict):
    """Single task in a PLAN_CREATED event."""

    skill_name: str
    reasoning: str


class PlanCreatedEvent(CustomEvent):
    """Event emitted after the planner creates an execution plan."""

    name: str = Field(default="PLAN_CREATED", frozen=True)
    value: list[PlanCreatedTaskValue]


def make_plan_created_event(tasks: list[PlanCreatedTaskValue]) -> PlanCreatedEvent:
    """Create a PLAN_CREATED event for yielding into the event stream."""
    return PlanCreatedEvent(timestamp=current_timestamp_ms(), value=tasks)


@dataclass
class RunContext:
    """Execution context passed to Planner and SkillRunner."""

    state: SearchState



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Annotated

from authlib.integrations.starlette_client import OAuth
from fastapi import Depends, HTTPException
from fastapi.security.http import HTTPAuthorizationCredentials
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.types import ASGIApp, Receive, Scope, Send
from starlette.websockets import WebSocket

from nwastdlib.url import URL
from oauth2_lib.fastapi import HTTPX_SSL_CONTEXT, AuthManager, HttpBearerExtractor, OIDCUserModel
from oauth2_lib.settings import oauth2lib_settings

oauth_client_credentials = OAuth()

oauth_client_credentials.register(
    "connext",
    server_metadata_url=URL(oauth2lib_settings.OIDC_CONF_URL),
    client_id=oauth2lib_settings.OAUTH2_RESOURCE_SERVER_ID,
    client_secret=oauth2lib_settings.OAUTH2_RESOURCE_SERVER_SECRET,
    request_token_params={"grant_type": "client_credentials"},
    client_kwargs={"verify": HTTPX_SSL_CONTEXT},
)

http_bearer_extractor = HttpBearerExtractor(auto_error=False)


async def authenticate(
    request: Request,
    http_auth_credentials: Annotated[HTTPAuthorizationCredentials | None, Depends(http_bearer_extractor)] = None,
) -> OIDCUserModel | None:
    token = http_auth_credentials.credentials if http_auth_credentials else None
    return await request.app.auth_manager.authentication.authenticate(request, token)


async def authorize(request: Request, user: Annotated[OIDCUserModel | None, Depends(authenticate)]) -> bool | None:
    return await request.app.auth_manager.authorization.authorize(request, user)


async def authenticate_websocket(websocket: WebSocket, token: str) -> OIDCUserModel | None:
    return await websocket.app.auth_manager.authentication.authenticate(websocket, token)


async def authorize_websocket(
    websocket: WebSocket, user: Annotated[OIDCUserModel | None, Depends(authenticate)]
) -> bool | None:
    return await websocket.app.auth_manager.authorization.authorize(websocket, user)


class AgentAuthMiddleware:
    """ASGI wrapper that enforces authn/authz on mounted agent sub-apps.

    Mounted sub-apps bypass FastAPI's dependency injection, so we enforce
    the same authenticate → authorize flow used by all other endpoints via
    a thin raw-ASGI layer that delegates to the parent app's auth_manager.
    """

    def __init__(self, app: ASGIApp, auth_manager: AuthManager) -> None:
        self.app = app
        self.auth_manager = auth_manager

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        if not oauth2lib_settings.OAUTH2_ACTIVE:
            await self.app(scope, receive, send)
            return

        headers = Headers(scope=scope)
        auth_header = headers.get("authorization", "")
        token = auth_header.removeprefix("Bearer ") if auth_header.startswith("Bearer ") else None

        request = Request(scope, receive)
        try:
            user = await self.auth_manager.authentication.authenticate(request, token)
            if user:
                await self.auth_manager.authorization.authorize(request, user)
            else:
                raise HTTPException(status_code=401, detail="Unauthorized")
        except HTTPException as exc:
            response = JSONResponse({"detail": exc.detail}, status_code=exc.status_code)
            await response(scope, receive, send)
            return

        await self.app(scope, receive, send)



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from uuid import UUID, uuid4

import structlog
from apscheduler.schedulers.base import BaseScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from sqlalchemy import delete

from orchestrator import app_settings
from orchestrator.db import db
from orchestrator.db.models import WorkflowApschedulerJob
from orchestrator.schemas.schedules import (
    APSchedulerJobCreate,
    APSchedulerJobDelete,
    APSchedulerJobs,
    APSchedulerJobUpdate,
    APSJobAdapter,
)
from orchestrator.services.processes import start_process
from orchestrator.services.workflows import get_workflow_by_workflow_id
from orchestrator.utils.errors import StartPredicateError
from orchestrator.utils.redis_client import create_redis_client

redis_connection = create_redis_client(app_settings.CACHE_URI.get_secret_value())

SCHEDULER_QUEUE = "scheduler:queue:"


logger = structlog.get_logger(__name__)


def serialize_payload(payload: APSchedulerJobs) -> bytes:
    """Serialize the payload to bytes for Redis storage.

    Args:
        payload: APSchedulerJobs The scheduled task payload.
    """
    data = json.loads(payload.model_dump_json())
    data["scheduled_type"] = payload.scheduled_type
    return json.dumps(data).encode()


def deserialize_payload(bytes_dump: bytes) -> APSchedulerJobs:
    """Deserialize the payload from bytes for Redis retrieval.

    Args:
        bytes_dump: bytes The serialized payload.
    """
    json_dump = bytes_dump.decode()
    return APSJobAdapter.validate_json(json_dump)


def add_scheduled_task_to_queue(payload: APSchedulerJobs) -> None:
    """Create a scheduled task service function.

    We need to create, update or delete an apscheduler job, and put the
    workflow and schedule_id in the linker table workflows_apscheduler_jobs.
    This is done by adding a job to a redis queue which will be executed
    when the scheduler runs.

    Args:
        payload: APSchedulerJobs The scheduled task to create, update or delete
    """
    bytes_dump = serialize_payload(payload)
    redis_connection.lpush(SCHEDULER_QUEUE, bytes_dump)
    logger.info("Added scheduled task to queue.")


def add_unique_scheduled_task_to_queue(payload: APSchedulerJobCreate) -> bool:
    """Create a unique scheduled task service function.

    Checks if the workflow is already scheduled before creating an apscheduler
    job, and putting the workflow and schedule_id in the linker table
    workflows_apscheduler_jobs.
    This is done by adding a job to a redis queue which will be executed
    when the scheduler runs.

    This function is not safe for concurrent usage and when the scheduler is not
    running, as there might be a race condition between adding a job and checking
    if it already exists in the database.

    Args:
        payload: APSchedulerJobCreate The scheduled task to create.

    Returns:
        True when the scheduled task was added to the queue
        False when the scheduled task was not added to the queue
    """
    if db.session.query(WorkflowApschedulerJob).filter_by(workflow_id=payload.workflow_id).all():
        logger.info(f"Not adding existing workflow {payload.workflow_name} as scheduled task.")
        return False
    add_scheduled_task_to_queue(payload)
    return True


def get_linker_entries_by_schedule_ids(schedule_ids: list[str]) -> list[WorkflowApschedulerJob]:
    """Get linker table entries for multiple schedule IDs in a single query.

    Args:
        schedule_ids: list[str] — One or many schedule IDs.

    Returns:
        list[WorkflowApschedulerJob]: All linker table rows matching those IDs.
    """
    if not schedule_ids:
        return []

    return db.session.query(WorkflowApschedulerJob).filter(WorkflowApschedulerJob.schedule_id.in_(schedule_ids)).all()


def _add_linker_entry(workflow_id: UUID, schedule_id: str) -> None:
    """Add an entry to the linker table workflows_apscheduler_jobs.

    Args:
        workflow_id: UUID The workflow ID.
        schedule_id: str The schedule ID.
    """
    workflows_apscheduler_job = WorkflowApschedulerJob(workflow_id=workflow_id, schedule_id=schedule_id)
    db.session.add(workflows_apscheduler_job)
    db.session.commit()


def _delete_linker_entry(workflow_id: UUID, schedule_id: str) -> None:
    """Delete an entry from the linker table workflows_apscheduler_jobs.

    Args:
        workflow_id: UUID The workflow ID.
        schedule_id: str The schedule ID.
    """
    db.session.execute(
        delete(WorkflowApschedulerJob).where(
            WorkflowApschedulerJob.workflow_id == workflow_id, WorkflowApschedulerJob.schedule_id == schedule_id
        )
    )
    db.session.commit()


def run_start_workflow_scheduler_task(workflow_name: str) -> None:
    """Function to start a workflow from the scheduler.

    Args:
        workflow_name: str The name of the workflow to start.
    """
    log = logger.bind(workflow_name=workflow_name)
    try:
        with db.database_scope():
            log.info("Starting workflow")
            process_id = start_process(workflow_name)
            log.info("Started workflow", process_id=process_id)
    except StartPredicateError:
        logger.info(f"Skipping {workflow_name} -> start predicate not satisfied")
    except Exception as e:
        log.exception(f"Failed to start {workflow_name} - unexpected error: {e}")


def _add_scheduled_task(payload: APSchedulerJobCreate, scheduler_connection: BaseScheduler) -> None:
    """Create a new scheduled task in the scheduler and also in the linker table.

    Args:
        payload: APSchedulerJobCreate The scheduled task to create.
        scheduler_connection: BaseScheduler The scheduler connection.
    """
    logger.info(f"Adding scheduled task: {payload}")

    workflow_description = None
    # Check if a workflow exists - we cannot schedule a non-existing workflow
    workflow = get_workflow_by_workflow_id(str(payload.workflow_id))
    if not workflow:
        raise ValueError(f"Workflow with id {payload.workflow_id} does not exist.")
    workflow_description = workflow.description

    # This function is always the same for scheduled tasks, it will run the workflow
    func = run_start_workflow_scheduler_task

    # Ensure payload has required data
    if not payload.trigger or not payload.workflow_name or not payload.trigger_kwargs or not payload.workflow_id:
        raise ValueError("Trigger must be specified for scheduled tasks.")

    schedule_id = str(uuid4())
    scheduler_connection.add_job(
        func=func,
        trigger=payload.trigger,
        id=schedule_id,
        name=payload.name or workflow_description,
        kwargs={"workflow_name": payload.workflow_name},
        **(payload.trigger_kwargs or {}),
    )

    _add_linker_entry(workflow_id=payload.workflow_id, schedule_id=schedule_id)


def _build_trigger_on_update(
    trigger_name: str | None, trigger_kwargs: dict
) -> IntervalTrigger | CronTrigger | DateTrigger | None:
    if not trigger_name or not trigger_kwargs:
        logger.info("Skipping building trigger as no trigger information is provided.")
        return None

    match trigger_name:
        case "interval":
            return IntervalTrigger(**trigger_kwargs)
        case "cron":
            return CronTrigger(**trigger_kwargs)
        case "date":
            return DateTrigger(**trigger_kwargs)
        case _:
            raise ValueError(f"Invalid trigger type: {trigger_name}")


def _update_scheduled_task(payload: APSchedulerJobUpdate, scheduler_connection: BaseScheduler) -> None:
    """Update an existing scheduled task in the scheduler.

    Only allow update of name and trigger
    Job id must be that of an existing job
    Do not insert in linker table - it should already exist.

    Args:
        payload: APSchedulerJobUpdate The scheduled task to update.
        scheduler_connection: BaseScheduler The scheduler connection.
    """
    logger.info(f"Updating scheduled task: {payload}")

    schedule_id = str(payload.schedule_id)
    job = scheduler_connection.get_job(job_id=schedule_id)
    if not job:
        raise ValueError(f"Schedule Job with id {schedule_id} does not exist.")

    trigger = _build_trigger_on_update(payload.trigger, payload.trigger_kwargs or {})
    modify_kwargs = {}

    if trigger:
        job = job.reschedule(trigger=trigger)

    if payload.name:
        modify_kwargs["name"] = payload.name

    job.modify(**modify_kwargs)


def _delete_scheduled_task(payload: APSchedulerJobDelete, scheduler_connection: BaseScheduler) -> None:
    """Delete an existing scheduled task in the scheduler and also in the linker table.

    Args:
        payload: APSchedulerJobDelete The scheduled task to delete.
        scheduler_connection: BaseScheduler The scheduler connection.
    """
    logger.info(f"Deleting scheduled task: {payload}")

    schedule_id = str(payload.schedule_id)
    scheduler_connection.remove_job(job_id=schedule_id)
    _delete_linker_entry(workflow_id=payload.workflow_id, schedule_id=schedule_id)


def workflow_scheduler_queue(queue_item: tuple[str, bytes], scheduler_connection: BaseScheduler) -> None:
    """Process an item from the scheduler queue.

    Args:
        queue_item: tuple[str, bytes] The item from the scheduler queue.
        scheduler_connection: BaseScheduler The scheduler connection.
    """
    try:
        _, bytes_dump = queue_item
        payload = deserialize_payload(bytes_dump)
        match payload:
            case APSchedulerJobCreate():
                _add_scheduled_task(payload, scheduler_connection)

            case APSchedulerJobUpdate():
                _update_scheduled_task(payload, scheduler_connection)

            case APSchedulerJobDelete():
                _delete_scheduled_task(payload, scheduler_connection)

            case _:
                logger.warning(f"Unexpected schedule type: {payload}")  # type: ignore
    except Exception:
        logger.exception("Error processing scheduler queue item")



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypedDict

from sqlalchemy import BindParameter, Select, and_, bindparam, case, cast, func, literal, or_, select
from sqlalchemy.sql.expression import ColumnElement, Label
from sqlalchemy.types import TypeEngine

from orchestrator.db.models import AiSearchIndex
from orchestrator.search.core.types import SearchMetadata

from ..pagination import PageCursor
from .base import Retriever


class RrfScoreSqlComponents(TypedDict):
    """SQL expression components of the RRF hybrid score calculation."""

    rrf_num: ColumnElement
    perfect: Label
    beta: ColumnElement
    rrf_max: ColumnElement
    fused_num: ColumnElement
    normalized_score: ColumnElement


def compute_rrf_hybrid_score_sql(
    sem_rank_col: ColumnElement,
    fuzzy_rank_col: ColumnElement,
    avg_fuzzy_score_col: ColumnElement,
    k: int,
    perfect_threshold: float,
    n_sources: int = 2,
    margin_factor: float = 0.05,
    score_numeric_type: TypeEngine | None = None,
) -> RrfScoreSqlComponents:
    """Compute RRF (Reciprocal Rank Fusion) hybrid score as SQL expressions for database execution.

    This function implements the core scoring logic for hybrid search combining semantic
    and fuzzy ranking. It computes:
    1. Base RRF score from both ranks
    2. Perfect match detection and boosting
    3. Dynamic beta parameter based on k and n_sources
    4. Normalized final score in [0, 1] range

    Args:
        sem_rank_col: SQLAlchemy column expression for semantic rank
        fuzzy_rank_col: SQLAlchemy column expression for fuzzy rank
        avg_fuzzy_score_col: SQLAlchemy column expression for average fuzzy score
        k: RRF constant controlling rank influence (typically 60)
        perfect_threshold: Threshold for perfect match boost (typically 0.9)
        n_sources: Number of ranking sources being fused (default: 2 for semantic + fuzzy)
        margin_factor: Margin above rrf_max as fraction (default: 0.05 = 5%)
        score_numeric_type: SQLAlchemy numeric type for casting scores

    Returns:
        RrfScoreSqlComponents: Dictionary of SQL expressions for score components
            - rrf_num: Raw RRF score (cast to numeric type if provided)
            - perfect: Perfect match flag (1 if avg_fuzzy_score >= threshold, else 0)
            - beta: Boost amount for perfect matches
            - rrf_max: Maximum possible RRF score
            - fused_num: RRF + perfect boost
            - normalized_score: Final score normalized to [0, 1]

    Note:
        -   Keep margin_factor small to avoid compressing perfects near 1 after normalization.

        -   The `beta` boost is calculated to be greater than the maximum possible standard
            RRF score (`rrf_max`). This guarantees that any item flagged as a "perfect" match
            will always rank above any non-perfect match.

        -   This function assumes that rank columns do not
            contain `NULL` values. A `NULL` in any rank column will result in a `NULL` final score
            for that item.
    """
    # RRF (rank-based): sum of 1/(k + rank_i) for each ranking source
    rrf_raw = (1.0 / (k + sem_rank_col)) + (1.0 / (k + fuzzy_rank_col))
    rrf_num = cast(rrf_raw, score_numeric_type) if score_numeric_type else rrf_raw

    # Perfect flag to boost near perfect fuzzy matches
    perfect = case((avg_fuzzy_score_col >= perfect_threshold, 1), else_=0).label("perfect_match")

    # Dynamic beta based on k and number of sources
    # rrf_max = n_sources / (k + 1)
    k_num = literal(float(k), type_=score_numeric_type) if score_numeric_type else literal(float(k))
    n_sources_lit = (
        literal(float(n_sources), type_=score_numeric_type) if score_numeric_type else literal(float(n_sources))
    )
    rrf_max = n_sources_lit / (k_num + literal(1.0, type_=score_numeric_type if score_numeric_type else None))

    margin = rrf_max * literal(margin_factor, type_=score_numeric_type if score_numeric_type else None)
    beta = rrf_max + margin

    # Fused score: RRF + perfect match boost
    perfect_casted = cast(perfect, score_numeric_type) if score_numeric_type else perfect
    fused_num = rrf_num + beta * perfect_casted

    # Normalize to [0,1] via the theoretical max (beta + rrf_max)
    norm_den = beta + rrf_max
    normalized_score = fused_num / norm_den

    return RrfScoreSqlComponents(
        rrf_num=rrf_num,
        perfect=perfect,
        beta=beta,
        rrf_max=rrf_max,
        fused_num=fused_num,
        normalized_score=normalized_score,
    )


class RrfHybridRetriever(Retriever):
    """Reciprocal Rank Fusion of semantic and fuzzy ranking with parent-child retrieval."""

    def __init__(
        self,
        q_vec: list[float],
        fuzzy_term: str,
        cursor: PageCursor | None,
        k: int = 60,
        field_candidates_limit: int = 100,
    ) -> None:
        self.q_vec = q_vec
        self.fuzzy_term = fuzzy_term
        self.cursor = cursor
        self.k = k
        self.field_candidates_limit = field_candidates_limit

    def apply(self, candidate_query: Select) -> Select:
        cand = candidate_query.subquery()
        q_param: BindParameter[list[float]] = bindparam("q_vec", self.q_vec, type_=AiSearchIndex.embedding.type)

        best_similarity = func.word_similarity(self.fuzzy_term, AiSearchIndex.value)
        sem_expr = case(
            (AiSearchIndex.embedding.is_(None), None),
            else_=AiSearchIndex.embedding.op("<->")(q_param),
        )
        sem_val = func.coalesce(sem_expr, literal(1.0)).label("semantic_distance")

        filter_condition = literal(self.fuzzy_term).op("<%")(AiSearchIndex.value)

        field_candidates = (
            select(
                AiSearchIndex.entity_id,
                AiSearchIndex.entity_title,
                AiSearchIndex.path,
                AiSearchIndex.value,
                sem_val,
                best_similarity.label("fuzzy_score"),
            )
            .select_from(AiSearchIndex)
            .join(cand, cand.c.entity_id == AiSearchIndex.entity_id)
            .where(
                and_(
                    AiSearchIndex.value_type.in_(self.SEARCHABLE_FIELD_TYPES),
                    filter_condition,
                )
            )
            .order_by(
                best_similarity.desc().nulls_last(),
                sem_expr.asc().nulls_last(),
                AiSearchIndex.entity_id.asc(),
            )
            .limit(self.field_candidates_limit)
        ).cte("field_candidates")

        entity_scores = (
            select(
                field_candidates.c.entity_id,
                field_candidates.c.entity_title,
                func.avg(field_candidates.c.semantic_distance).label("avg_semantic_distance"),
                func.avg(field_candidates.c.fuzzy_score).label("avg_fuzzy_score"),
            ).group_by(field_candidates.c.entity_id, field_candidates.c.entity_title)
        ).cte("entity_scores")

        entity_highlights = (
            select(
                field_candidates.c.entity_id,
                func.first_value(field_candidates.c.value)
                .over(
                    partition_by=field_candidates.c.entity_id,
                    order_by=[field_candidates.c.fuzzy_score.desc(), field_candidates.c.path.asc()],
                )
                .label(self.HIGHLIGHT_TEXT_LABEL),
                func.first_value(field_candidates.c.path)
                .over(
                    partition_by=field_candidates.c.entity_id,
                    order_by=[field_candidates.c.fuzzy_score.desc(), field_candidates.c.path.asc()],
                )
                .label(self.HIGHLIGHT_PATH_LABEL),
            ).distinct(field_candidates.c.entity_id)
        ).cte("entity_highlights")

        ranked = (
            select(
                entity_scores.c.entity_id,
                entity_scores.c.entity_title,
                entity_scores.c.avg_semantic_distance,
                entity_scores.c.avg_fuzzy_score,
                entity_highlights.c.highlight_text,
                entity_highlights.c.highlight_path,
                func.dense_rank()
                .over(
                    order_by=[entity_scores.c.avg_semantic_distance.asc().nulls_last(), entity_scores.c.entity_id.asc()]
                )
                .label("sem_rank"),
                func.dense_rank()
                .over(order_by=[entity_scores.c.avg_fuzzy_score.desc().nulls_last(), entity_scores.c.entity_id.asc()])
                .label("fuzzy_rank"),
            ).select_from(
                entity_scores.join(entity_highlights, entity_scores.c.entity_id == entity_highlights.c.entity_id)
            )
        ).cte("ranked_results")

        # Compute RRF hybrid score
        score_components = compute_rrf_hybrid_score_sql(
            sem_rank_col=ranked.c.sem_rank,
            fuzzy_rank_col=ranked.c.fuzzy_rank,
            avg_fuzzy_score_col=ranked.c.avg_fuzzy_score,
            k=self.k,
            perfect_threshold=0.9,
            score_numeric_type=self.SCORE_NUMERIC_TYPE,
        )

        perfect = score_components["perfect"]
        normalized_score = score_components["normalized_score"]

        # Round to configured precision
        score = cast(
            func.round(cast(normalized_score, self.SCORE_NUMERIC_TYPE), self.SCORE_PRECISION),
            self.SCORE_NUMERIC_TYPE,
        ).label(self.SCORE_LABEL)

        stmt = select(
            ranked.c.entity_id,
            ranked.c.entity_title,
            score,
            ranked.c.highlight_text,
            ranked.c.highlight_path,
            perfect.label("perfect_match"),
        ).select_from(ranked)

        stmt = self._apply_fused_pagination(stmt, score, ranked.c.entity_id)

        return stmt.order_by(
            score.desc().nulls_last(),
            ranked.c.entity_id.asc(),
        ).params(q_vec=self.q_vec)

    def _apply_fused_pagination(
        self,
        stmt: Select,
        score_column: ColumnElement,
        entity_id_column: ColumnElement,
    ) -> Select:
        """Keyset paginate by fused score + id."""
        if self.cursor is not None:
            score_param = self._quantize_score_for_pagination(self.cursor.score)
            stmt = stmt.where(
                or_(
                    score_column < score_param,
                    and_(score_column == score_param, entity_id_column > self.cursor.id),
                )
            )
        return stmt

    @property
    def metadata(self) -> SearchMetadata:
        return SearchMetadata.hybrid()



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy import select, text
from sqlalchemy.exc import ProgrammingError
from sqlalchemy_utils import Ltree

from orchestrator.db import db
from orchestrator.db.database import WrappedSession
from orchestrator.db.models import AiSearchIndex
from orchestrator.schemas.search_requests import SearchRequest
from orchestrator.search.aggregations import AggregationType
from orchestrator.search.core.types import EntityType, FieldType
from orchestrator.search.filters import FilterCondition, FilterTree, LtreeFilter, PathFilter
from orchestrator.search.filters.definitions import operators_for
from orchestrator.search.query.exceptions import (
    EmptyFilterPathError,
    IncompatibleAggregationTypeError,
    IncompatibleFilterTypeError,
    IncompatibleTemporalGroupingTypeError,
    InvalidEntityPrefixError,
    InvalidLtreePatternError,
    PathNotFoundError,
)
from orchestrator.search.query.mixins import OrderBy


def is_filter_compatible_with_field_type(filter_condition: FilterCondition, field_type: FieldType) -> bool:
    """Check whether a filter condition is compatible with a given field type.

    Args:
        filter_condition (FilterCondition): The filter condition instance to check.
        field_type (FieldType): The type of field from the index schema.

    Returns:
        bool: True if the filter condition is valid for the given field type, False otherwise.
    """

    # LtreeFilter is for path filtering only and is thus compatible with all field types.
    if isinstance(filter_condition, LtreeFilter):
        return True

    # Get valid operators for this field type and check if the filter's operator is valid.
    valid_operators = operators_for(field_type)
    return filter_condition.op in valid_operators


def is_lquery_syntactically_valid(pattern: str, db_session: WrappedSession) -> bool:
    """Validate whether a string is a syntactically correct `lquery` pattern.

    Args:
        pattern (str): The LTree lquery pattern string to validate.
        db_session (WrappedSession): The database session used to test casting.

    Returns:
        bool: True if the pattern is valid, False if it fails to cast in PostgreSQL.
    """

    try:
        with db_session.begin_nested():
            db_session.execute(text("SELECT CAST(:pattern AS lquery)"), {"pattern": pattern})
        return True
    except ProgrammingError:
        return False


def get_structured_filter_schema() -> dict[str, str]:
    """Retrieve all distinct filterable paths and their field types from the index.

    Returns:
        Dict[str, str]: Mapping of path strings to their corresponding field type values.
    """

    stmt = select(AiSearchIndex.path, AiSearchIndex.value_type).distinct().order_by(AiSearchIndex.path)
    result = db.session.execute(stmt)
    return {str(path): value_type.value for path, value_type in result}


def validate_filter_path(path: str) -> str | None:
    """Check if a given path exists in the index and return its field type.

    Args:
        path (str): The fully qualified LTree path.

    Returns:
        Optional[str]: The value type of the field if found, otherwise None.
    """

    stmt = select(AiSearchIndex.value_type).where(AiSearchIndex.path == Ltree(path)).limit(1)
    result = db.session.execute(stmt).scalar_one_or_none()
    return result.value if result else None


async def complete_filter_validation(filter: PathFilter, entity_type: EntityType) -> None:
    """Validate a PathFilter against the database schema and entity type.

    Checks performed:
    1. LTree filter syntax (for LtreeFilter only)
    2. Non-empty path
    3. Path exists in the database schema
    4. Filter type matches the field's value_type
    5. Path starts with the correct entity type prefix (unless wildcard)

    Args:
        filter (PathFilter): The filter to validate.
        entity_type (EntityType): The entity type being searched.

    Raises:
        ValueError: If any of the validation checks fail.
    """

    # Ltree is a special case
    if isinstance(filter.condition, LtreeFilter):
        lquery_pattern = filter.condition.value
        if not is_lquery_syntactically_valid(lquery_pattern, db.session):
            raise InvalidLtreePatternError(lquery_pattern)
        return

    if not filter.path or not filter.path.strip():
        raise EmptyFilterPathError()

    # 1. Check if path exists in database
    db_field_type_str = validate_filter_path(filter.path)
    if db_field_type_str is None:
        raise PathNotFoundError(filter.path)

    db_field_type = FieldType(db_field_type_str)

    # 2. Check filter compatibility with field type
    if not is_filter_compatible_with_field_type(filter.condition, db_field_type):
        expected_operators = operators_for(db_field_type)
        raise IncompatibleFilterTypeError(
            filter.condition.op.value, db_field_type.value, filter.path, expected_operators
        )

    # 3. Check entity type prefix requirements (unless it's a wildcard path)
    expected_prefix = f"{entity_type.value.lower()}."
    if not filter.path.startswith(expected_prefix) and not filter.path.startswith("*"):
        raise InvalidEntityPrefixError(filter.path, expected_prefix, entity_type.value)


async def validate_filter_tree(filters: FilterTree | None, entity_type: EntityType) -> None:
    """Validate all PathFilter leaves in a FilterTree."""
    if filters is None:
        return
    for leaf in filters.get_all_leaves():
        await complete_filter_validation(leaf, entity_type)


def validate_aggregation_field(agg_type: AggregationType, field_path: str) -> None:
    """Validate that an aggregation field exists and is compatible with the aggregation type.

    Note: Only for FieldAggregations (SUM, AVG, MIN, MAX). COUNT does not require field validation.

    Args:
        agg_type: The aggregation type enum
        field_path: The field path to validate

    Raises:
        PathNotFoundError: If the field doesn't exist in the database.
        IncompatibleAggregationTypeError: If the field type is incompatible with the aggregation type.
    """
    # Check if field exists in database
    field_type_str = validate_filter_path(field_path)
    if field_type_str is None:
        raise PathNotFoundError(field_path)

    # Validate field type compatibility with aggregation type
    if agg_type in (AggregationType.SUM, AggregationType.AVG):
        if field_type_str not in (FieldType.INTEGER.value, FieldType.FLOAT.value):
            raise IncompatibleAggregationTypeError(
                agg_type.value, field_type_str, field_path, [FieldType.INTEGER.value, FieldType.FLOAT.value]
            )
    elif agg_type in (AggregationType.MIN, AggregationType.MAX):
        if field_type_str not in (FieldType.INTEGER.value, FieldType.FLOAT.value, FieldType.DATETIME.value):
            raise IncompatibleAggregationTypeError(
                agg_type.value,
                field_type_str,
                field_path,
                [FieldType.INTEGER.value, FieldType.FLOAT.value, FieldType.DATETIME.value],
            )


def validate_temporal_grouping_field(field_path: str) -> None:
    """Validate that a field exists and is a datetime type for temporal grouping.

    Args:
        field_path: The field path to validate

    Raises:
        PathNotFoundError: If the field doesn't exist in the database
        IncompatibleTemporalGroupingTypeError: If the field is not a datetime type
    """
    # Check if field exists in database
    field_type_str = validate_filter_path(field_path)
    if field_type_str is None:
        raise PathNotFoundError(field_path)

    # Validate field type is datetime
    if field_type_str != FieldType.DATETIME.value:
        raise IncompatibleTemporalGroupingTypeError(field_path, field_type_str)


def validate_grouping_fields(group_by_paths: list[str]) -> None:
    """Validate that all grouping field paths exist in the database.

    Args:
        group_by_paths: List of field paths to group by

    Raises:
        PathNotFoundError: If any path doesn't exist in the database
    """
    for path in group_by_paths:
        field_type = validate_filter_path(path)
        if field_type is None:
            raise PathNotFoundError(path)


def validate_order_by_fields(order_by: list[OrderBy] | None) -> None:
    """Validate that order_by field paths exist in the database.

    Args:
        order_by: List of ordering instructions, or None

    Raises:
        PathNotFoundError: If a field path doesn't exist in the database

    Note:
        Only validates fields that appear to be paths (contain dots).
        Aggregation aliases (no dots, like 'count') are skipped as they
        cannot be validated until query execution time.
    """
    if order_by is None:
        return

    for order_instr in order_by:
        # Skip aggregation aliases (no dots, e.g., 'count', 'revenue')
        if "." not in order_instr.field:
            continue

        field_type = validate_filter_path(order_instr.field)
        if field_type is None:
            raise PathNotFoundError(order_instr.field)


def get_ai_search_index_by_entity_type_and_path(entity_type: EntityType, path: str) -> AiSearchIndex | None:
    stmt = (
        select(AiSearchIndex)
        .where(AiSearchIndex.path == Ltree(path), AiSearchIndex.entity_type == entity_type.value)
        .limit(1)
    )
    return db.session.execute(stmt).scalar_one_or_none()


def validate_structured_order_by_element(entity_type: EntityType | None, request: SearchRequest | None) -> None:
    if request and request.order_by and entity_type:
        element = request.order_by.element
        exists = get_ai_search_index_by_entity_type_and_path(entity_type, element)
        if not exists:
            raise ValueError(f"Element {element} is not a valid path")



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, cast

import structlog
from pydantic_ai import RunContext
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.exceptions import ModelRetry
from pydantic_ai.toolsets import FunctionToolset

from orchestrator.api.api_v1.endpoints.search import (
    get_definitions,
    list_paths,
)
from orchestrator.search.agent.state import SearchState
from orchestrator.search.core.types import EntityType, FilterOp, QueryOperation
from orchestrator.search.filters import FilterTree
from orchestrator.search.query.exceptions import PathNotFoundError, QueryValidationError
from orchestrator.search.query.queries import AggregateQuery, CountQuery, Query, SelectQuery
from orchestrator.search.query.validation import validate_filter_tree

logger = structlog.get_logger(__name__)


def ensure_query_initialized(
    state: SearchState,
    entity_type: EntityType,
    query_operation: QueryOperation = QueryOperation.SELECT,
) -> None:
    """Lazy-initialize query on state if not already set, or re-create if type mismatches.

    Preserves filters (from existing query or pending_filters) when creating/upgrading.
    """
    expected_types = {
        QueryOperation.SELECT: SelectQuery,
        QueryOperation.COUNT: CountQuery,
        QueryOperation.AGGREGATE: AggregateQuery,
    }
    expected_type = expected_types[query_operation]

    if state.query is not None and isinstance(state.query, expected_type):
        return

    if state.query is not None:
        logger.warning(
            "Query type mismatch — re-creating query, grouping/aggregations will be lost",
            existing_type=type(state.query).__name__,
            requested_operation=query_operation.value,
        )

    # Collect filters: from existing query, or from pending_filters set by set_filter_tree
    filters = (state.query.filters if state.query else None) or state.pending_filters

    if query_operation == QueryOperation.SELECT:
        state.query = SelectQuery(entity_type=entity_type, query_text=state.user_input, filters=filters)
    elif query_operation == QueryOperation.COUNT:
        state.query = CountQuery(entity_type=entity_type, filters=filters)
    else:
        state.query = AggregateQuery(entity_type=entity_type, aggregations=[], filters=filters)

    state.pending_filters = None


filter_building_toolset: FunctionToolset[StateDeps[SearchState]] = FunctionToolset(max_retries=2)


@filter_building_toolset.tool
async def set_filter_tree(
    ctx: RunContext[StateDeps[SearchState]],
    filters: FilterTree | None,
    entity_type: EntityType,
) -> FilterTree | None:
    """Replace current filters atomically with a full FilterTree, or clear with None.

    See FilterTree model for structure, operators, and examples.
    Filters are validated immediately and applied when the query executes.
    """
    logger.debug(
        "Setting filter tree",
        entity_type=entity_type.value,
        has_filters=filters is not None,
        filter_summary=f"{len(filters.get_all_leaves())} filters" if filters else "no filters",
    )

    try:
        await validate_filter_tree(filters, entity_type)
    except PathNotFoundError as e:
        logger.debug(f"{PathNotFoundError.__name__}: {str(e)}")
        raise ModelRetry(f"{str(e)} Use discover_filter_paths tool to find valid paths.")
    except QueryValidationError as e:
        logger.debug(f"Query validation failed: {str(e)}")
        raise ModelRetry(str(e))
    except Exception as e:
        logger.error("Unexpected Filter validation exception", error=str(e))
        raise ModelRetry(f"Filter validation failed: {str(e)}. Please check your filter structure and try again.")

    # Store validated filters — applied when query is created by run_search/run_aggregation
    if ctx.deps.state.query is not None:
        ctx.deps.state.query = cast(Query, ctx.deps.state.query).model_copy(update={"filters": filters})
    else:
        ctx.deps.state.pending_filters = filters

    return filters


@filter_building_toolset.tool
async def discover_filter_paths(
    ctx: RunContext[StateDeps[SearchState]],
    field_names: list[str],
    entity_type: EntityType | None = None,
) -> dict[str, dict[str, Any]]:
    """Discovers available filter paths for a list of field names.

    Returns a dictionary where each key is a field_name from the input list and
    the value is its discovery result.
    """
    if not entity_type:
        if ctx.deps.state.query:
            entity_type = ctx.deps.state.query.entity_type
        else:
            raise ModelRetry("Entity type not specified and no query in state. Pass entity_type.")

    all_results = {}
    for field_name in field_names:
        paths_response = await list_paths(prefix="", q=field_name, entity_type=entity_type, limit=100)

        matching_leaves = []
        for leaf in paths_response.leaves:
            if field_name.lower() in leaf.name.lower():
                matching_leaves.append(
                    {
                        "name": leaf.name,
                        "value_kind": leaf.ui_types,
                        "paths": leaf.paths,
                    }
                )

        matching_components = []
        for comp in paths_response.components:
            if field_name.lower() in comp.name.lower():
                matching_components.append(
                    {
                        "name": comp.name,
                        "value_kind": comp.ui_types,
                    }
                )

        result_for_field: dict[str, Any]
        if not matching_leaves and not matching_components:
            result_for_field = {
                "status": "NOT_FOUND",
                "guidance": f"No filterable paths found containing '{field_name}'. Do not create a filter for this.",
                "leaves": [],
                "components": [],
            }
        else:
            result_for_field = {
                "status": "OK",
                "guidance": f"Found {len(matching_leaves)} field(s) and {len(matching_components)} component(s) for '{field_name}'.",
                "leaves": matching_leaves,
                "components": matching_components,
            }

        all_results[field_name] = result_for_field
    return all_results


@filter_building_toolset.tool
async def get_valid_operators() -> dict[str, list[FilterOp]]:
    """Gets the mapping of field types to their valid filter operators."""
    definitions = await get_definitions()

    operator_map = {}
    for ui_type, type_def in definitions.items():
        key = ui_type.value

        if hasattr(type_def, "operators"):
            operator_map[key] = type_def.operators
    return operator_map
